/*
 *  jODIE  -- A Java Application for interacting with the OTIS utility "ODIE".
 *
 *  Copyright (C) 2013-2024, Joseph A. Huwaldt. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.jodie.AppGUI;
import jahuwaldt.jodie.AppPreferences;
import jahuwaldt.jodie.MainWindow;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.MainApp;
import jahuwaldt.swing.Preferences;
import java.io.File;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JOptionPane;

/**
 * An application that can interact with the OTIS utility ODIE (OTIS Data Interactive
 * Examiner) to understand how OTIS will interpolate data tables.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: November, 4, 2013
 * @version January 19, 2024
 */
public class jODIE implements MainApp {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(jODIE.class.getName());

    /**
     * The resource bundle containing the string resources for this application. Resource
     * bundles are useful for localizing applications. New localities (languages) can be
     * added by adding properties files only -- no code modifications.
     */
    public static final ResourceBundle RESBUNDLE = ResourceBundle.getBundle("appStrings", Locale.getDefault());
    
    //	A reference to our GUI application instance.
    private final MDIApplication guiApp;
    
    /**
     * The user preferences for this program.
     */
    private final AppPreferences prefs;
    

    //-----------------------------------------------------------------------------------
    /**
     * Entry point for this application. This is where it starts up from.
     * 
     * @param args Command line arguments.
     */
    public static void main(final String[] args) {

        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                createAndShowGUI(args);
            }
        });

    }

    /**
     * Create the GUI and show it. For thread safety, this method should be invoked from
     * the event-dispatching thread.
     */
    private static void createAndShowGUI(String[] args) {

        try {
            // Create an instance of our application.
            if (args.length > 0) {
                new jODIE(args);
            } else {
                new jODIE();
            }


        } catch (NoSuchMethodException e) {
            LOGGER.log(Level.SEVERE, null, e);
            AppUtilities.showException(null, RESBUNDLE.getString("noSuchMethodTile"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
            System.exit(0);

        } catch (Throwable e) {
            LOGGER.log(Level.SEVERE, null, e);
            AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
            System.exit(0);
        }
    }

    /**
     * Construct a new jODIE application and open and empty window.
     */
    public jODIE() throws NoSuchMethodException, Exception {

        guiApp = new AppGUI(RESBUNDLE);

        //	Load in the user preferences.
        prefs = new AppPreferences();
        MDIApplication.setPreferences(prefs);

        // Check the configuration.
        boolean setPrefs = checkPathPreference(AppPreferences.ODIEPATH, "ODIE");

        //  Create a main window frame for this application.
        MainWindow window = new MainWindow();
        window.setVisible(true);
        MDIApplication.addWindow(window);

        //	Display the preferences dialog if necessary.
        if (setPrefs)
            prefs.showPreferenceDialog();
    }

    /**
     * Construct a new jODIE application and load in the files provided on the command line.
     */
    public jODIE(String[] args) throws NoSuchMethodException, Exception {

        guiApp = new AppGUI(RESBUNDLE);

        //	Load in the user preferences.
        prefs = new AppPreferences();

        // Check the configuration.
        boolean setPrefs = checkPathPreference(AppPreferences.ODIEPATH, "ODIE");

        //	Deal with the command line arguments.

        //	Return a list of arguments without the options.
        List<String> argsNoOptions = argumentsWithoutOptions(args);

        if (argumentStartsWith(args, "-h") || argumentStartsWith(args, "--help"))
            printHelpAndDie();

        if (argumentStartsWith(args, "-v")) {
            System.out.print("jODIE version ");
            System.out.println(RESBUNDLE.getString("appVersion"));
            System.out.println();
        }

        //	Create an instance of the application window using the supplied data file.
        if (!argsNoOptions.isEmpty()) {
            List<File> files = new ArrayList<>();
            for (String filename : argsNoOptions) {
                File file = new File(filename);
                if (file.exists() && file.canRead())
                    files.add(file);
            }
            MainWindow.newWindowFromDataFiles(files);

        } else
            System.exit(0);


        //	Display the preferences dialog if necessary.
        if (setPrefs)
            prefs.showPreferenceDialog();
    }

    /**
     * Checks to see if the specified path exists in the preferences.
     *
     * @param prefKey The key to the preference path to look for.
     * @param fileDesc The brief description of the file being looked for.
     */
    private boolean checkPathPreference(String key, String fileDesc) {
        String path = prefs.get(key);
        if (path == null) {
            JOptionPane.showMessageDialog(null,
                    MessageFormat.format(RESBUNDLE.getString("execPathNotSet"), fileDesc));
            return true;
        }
        File file = new File(path);
        if (!file.exists()) {
            JOptionPane.showMessageDialog(null,
                    MessageFormat.format(RESBUNDLE.getString("execPathNotValid"), fileDesc));
            return true;
        }
        return false;
    }

    /**
     * Returns true if any of the arguments starts with the specified string (ignoring
     * case).
     */
    private boolean argumentStartsWith(String[] args, String value) {
        value = value.toLowerCase();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (args[i].toLowerCase().startsWith(value))
                return true;
        }
        return false;
    }

    /**
     * Takes the input list of arguments and strips out any arguments that are options
     * (i.e.: start with "-").
     */
    private List<String> argumentsWithoutOptions(String[] args) {
        List<String> list = new ArrayList();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (!args[i].startsWith("-"))
                list.add(args[i]);

            //	Skip the next argument if this one starts with "--".
            if (args[i].startsWith("--"))
                ++i;
        }
        return list;
    }

    /**
     * Return the index to the argument that starts with the specified value (ignoring
     * case).
     *
     * @param args The list of all the arguments.
     * @param value The value we are looking for.
     * @return the index to the argument that starts with the specified value ignoring
     *      case or -1 if none could be found.
     */
    private int argumentIndex(String[] args, String value) {
        value = value.toLowerCase();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (args[i].toLowerCase().startsWith(value))
                return i;
        }
        return -1;
    }

    /**
     * Method that prints a help message to standard out and then exits the program.
     */
    private static void printHelpAndDie() {
        System.out.println();
        System.out.println("jODIE -- A GUI interface for ODIE");
        System.out.println("Usage:");
        System.out.println("    java -jar jODIE.jar filename1 filename2 etc");
        System.out.println("    where:");
        System.out.println("        filename* = A series of table files to be loaded.");
        System.out.println();
        System.exit(0);
    }
    
    //-----------------------------------------------------------------------------------
    /**
     * Returns the resource bundle stored with this application. If no resource bundle has
     * been stored, then null is returned.
     */
    @Override
    public final ResourceBundle getResourceBundle() {
        return RESBUNDLE;
    }

    /**
     * Return a reference to this program's GUI application or null if we are running in
     * batch mode.
     */
    @Override
    public final MDIApplication getGUIApplication() {
        return guiApp;
    }

    /**
     * Return a reference to the user preferences for this application.
     */
    @Override
    public final Preferences getPreferences() {
        return prefs;
    }
}
