package jahuwaldt.jodie;

import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.io.StreamGobbler;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.EscapeJDialog;
import jahuwaldt.swing.MDIApplication;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.File;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.prefs.Preferences;
import javax.swing.*;

/**
 * This class serves as a collection of preferences for this program.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt Date: November 4, 2013
 * @version January 19, 2024
 */
public class AppPreferences implements jahuwaldt.swing.Preferences {

    //  The preferences for this application are stored as a properties list.
    private final Preferences prefs = Preferences.userNodeForPackage(this.getClass());
    
    //  Use a change-listener system to keep the preferences dialog updated
    private final PropertyChangeSupport propertyChange = new PropertyChangeSupport(this);
    
    //  Preference keys
    private static final String LASTPATH = "lastPath";
    
    /**
     * The preference key for the ODIE application executable path.
     */
    public static final String ODIEPATH = "odiePath";
    
    private static final String ODIEPATH_DEFAULT = "odie";
    
    //  The best guess at the path to the required applications.
    private static String odiePath_default = ODIEPATH_DEFAULT;
    static {
        if (!AppUtilities.isWindows()) {
            //  Assume we have a UNIX-like system.
            odiePath_default = findExecutablePath("odie", ODIEPATH_DEFAULT);
        }
    }
    
    //  A reference to the window containing the application preferences.
    private PrefsDialog prefsDialog = null;
    

    /**
     * Construct the preferences object for this application. This constructor will locate
     * the preference file and load in any available preferences for the application.
     *
     * @param app Reference to the main application.
     */
    public AppPreferences() {
        //  Do nothing.
    }

    /**
     * Returns the file path to the last referenced file. Returns null if
     * no last path could be found.
     */
    @Override
    public String getLastPath() {
        return prefs.get(LASTPATH, null);
    }

    /**
     * Set the last file path referenced by the user. This is the path to the last
     * referenced file.
     */
    @Override
    public void setLastPath(String path) {
        String oldPath = prefs.get(LASTPATH, null);
        if ((path != null && !path.equals(oldPath)) || (path == null && oldPath != null)) {
            prefs.put(LASTPATH, path);
        }
    }

    /**
     * Return the path to the ODIE application.
     */
    public String getODIEPath() {
        return prefs.get(ODIEPATH, odiePath_default);
    }

    /**
     * Set the path to the ODIE application.
     */
    public synchronized void setODIEPath(String path) {
        String oldPath = prefs.get(ODIEPATH, null);
        if ((path != null && !path.equals(oldPath)) || (path == null && oldPath != null)) {
            prefs.put(ODIEPATH, path);
        }
    }

    /**
     * Return the preference with the specified key String.
     *
     * @param key The key String identifying the preference to be retrieved.
     */
    @Override
    public String get(String key) {
        return prefs.get(key, null);
    }

    /**
     * Set the preference with the specified key String.
     *
     * @param key The key String identifying the preference to be set.
     * @param value THe String value to store as the preference.
     */
    @Override
    public void set(String key, String value) {
        prefs.put(key, value);
    }

    public synchronized void addPropertyChangeListener(PropertyChangeListener l) {
        propertyChange.addPropertyChangeListener(l);
    }

    public synchronized void addPropertyChangeListener(String propertyName, PropertyChangeListener l) {
        propertyChange.addPropertyChangeListener(propertyName, l);
    }

    public synchronized void removePropertyChangeListener(PropertyChangeListener l) {
        propertyChange.removePropertyChangeListener(l);
    }

    public synchronized void removePropertyChangeListener(String propertyName, PropertyChangeListener l) {
        propertyChange.removePropertyChangeListener(propertyName, l);
    }

    /**
     * Method that displays a dialog that allows the user to change the application
     * preferences. If the application is running in batch mode, this method does nothing.
     */
    @Override
    public void showPreferenceDialog() {
        if (prefsDialog != null)
            prefsDialog.setVisible(true);
        else {
            try {

                prefsDialog = new PrefsDialog();
                prefsDialog.pack();
                AppUtilities.positionWindow(prefsDialog, prefsDialog.getWidth(), prefsDialog.getHeight());
                prefsDialog.setVisible(true);

            } catch (NoSuchMethodException e) {
                e.printStackTrace();	//	Should only happen during development.
            }
        }
    }

    /**
     * This dialog used to allow the user to set the preferences for this program.
     */
    private class PrefsDialog extends EscapeJDialog {

        @SuppressWarnings("OverridableMethodCallInConstructor")
        public PrefsDialog() throws NoSuchMethodException {
            super(null, MDIApplication.getInstance().getResourceBundle().getString("prefsPanelTitle"), false);
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

            //  Layout the dialog window.
            Container cp = this.getContentPane();
            cp.setLayout(new BorderLayout());

            //	Add a close button to the bottom.
            JButton close = new JButton(resBundle.getString("closeItemText"));
            close.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    PrefsDialog.this.setVisible(false);
                }
            });
            Box box = Box.createHorizontalBox();
            box.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
            box.add(Box.createHorizontalGlue());
            box.add(close);
            cp.add(box, BorderLayout.SOUTH);
            this.getRootPane().setDefaultButton(close);	//	Make the default button.

            //	Add executable paths.
            JPanel panel = new JPanel();
            panel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10),
                    BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.black),
                    resBundle.getString("execPathsLabel"))));
            cp.add(panel, BorderLayout.NORTH);

            //	Add the ODIE executable path.
            createFileInput(ODIEPATH, odiePath_default, resBundle.getString("odieExecLabel"), 30,
                    resBundle.getString("odieExecToolTipText"), panel);

        }

        /**
         * Construct a new file selection JTextField with an associated label and
         * selection button and add these three items to the specified panel.
         *
         * @param prefKey The preferences database key used by this file input field.
         * @param defPath The default path to the file referred to be prefKey.
         * @param label The label used for this input field.
         * @param columns The number of columns for the text field.
         * @param btnToolTip The tool-tip to show for the select button.
         * @param panel The panel that the text field and related items are added to.
         */
        private JTextField createFileInput(final String prefKey, String defPath, String label, int columns, final String btnToolTip,
                JPanel panel) {
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

            panel.add(new JLabel(label, SwingConstants.RIGHT));
            final JTextField tf = new JTextField(columns);
            tf.setEditable(true);
            tf.setDragEnabled(true);
            tf.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent evt) {
                    JTextField tf = (JTextField) evt.getSource();
                    String path = tf.getText();
                    File file = new File(path);
                    if (file.exists())
                        AppPreferences.this.set(prefKey, path);
                    else {
                        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                        JOptionPane.showMessageDialog(PrefsDialog.this,
                                MessageFormat.format(resBundle.getString("pathDoesntExit"), prefKey),
                                resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);
                    }
                }
            });
            tf.addFocusListener(new FocusAdapter() {
                @Override
                public void focusLost(FocusEvent evt) {
                    JTextField tf = (JTextField) evt.getSource();
                    String oldPath = AppPreferences.this.get(prefKey);
                    String path = tf.getText();
                    File file = new File(path);
                    if (file.exists())
                        AppPreferences.this.set(prefKey, path);
                    else if (oldPath != null && !path.equals(oldPath)) {
                        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                        JOptionPane.showMessageDialog(PrefsDialog.this,
                                MessageFormat.format(resBundle.getString("pathDoesntExit"), prefKey),
                                resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

            String path = AppPreferences.this.get(prefKey);
            if (path == null)
                path = defPath;
            tf.setText(path);
            tf.setCaretPosition(tf.getDocument().getLength());

            panel.add(tf);
            JButton btn = new JButton(resBundle.getString("browseBtnTxt"));
            btn.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doChooseFile(prefKey, tf, btnToolTip, null);
                }
            });
            btn.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
            btn.setToolTipText(btnToolTip);
            panel.add(btn);

            return tf;
        }

        /**
         * Ask the user to select a file using a file chooser dialog. Store the results of
         * the file selection in the specified preferences key and text field.
         *
         * @param pefKey The preferences database key where the file path is to be stored.
         * @param tf The text field that is to be filled in with the selected path.
         * @param msg The message to display to the user in the file chooser dialog.
         * @param filter The file name filter to use or <code>null</code> for none.
         */
        public void doChooseFile(String prefKey, JTextField tf, String msg, ExtFilenameFilter filter) {
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

            try {
                String path = AppPreferences.this.get(prefKey);
                String parent = null, name = null;
                if (path != null) {
                    File theFile = new File(path);
                    parent = theFile.getParent();
                    name = theFile.getName();
                }
                File theFile = AppUtilities.selectFile(prefsDialog, FileDialog.LOAD, msg,
                        parent, name, filter);
                if (theFile == null)
                    return;

                String newPath = theFile.getPath();
                if (!newPath.equals(path)) {
                    //	Set the project path text field.
                    tf.setText(newPath);

                    //	Store off the path reference for later use.
                    AppPreferences.this.set(prefKey, newPath);
                }

            } catch (Exception e) {
                AppUtilities.showException(prefsDialog, resBundle.getString("unexpectedTitle"),
                        resBundle.getString("unexpectedMsg"), e);
                e.printStackTrace();
            }
        }

        //	Add support for Control-W to close the window.
        @Override
        public void keyPressed(KeyEvent e) {
            super.keyPressed(e);

            int accCharMask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMaskEx();
            KeyStroke ks1 = KeyStroke.getKeyStroke('W', accCharMask, false);

            int keyCode = e.getKeyCode();
            int modifiers = e.getModifiersEx();
            KeyStroke ks2 = KeyStroke.getKeyStroke(keyCode, modifiers);

            if (ks1.equals(ks2))
                setVisible(false);

        }
    }

    /**
     * Method that executes the specified command as a separate process and returns a list
     * containing the output to standard out and standard error for that process.
     *
     * @param cmds The array of commands and arguments to run in the system's command
     *      shell (Unix "sh" or DOS cmd.exe shell).
     * @param dir The working directory for the process to run in.
     * @param wait Wait for the command to complete before returning if true. Return after
     *      a short pause if false.
     * @return A list of strings that represent the combination of standard out and
     *      standard error from the process.
     * @throws IOException if there is any I/O error.
     * @throws InterruptedException if the current thread is interrupted by another thread
     *      while it is waiting.
     */
    private static List<String> runCommand(List<String> cmds, File dir, boolean wait) throws IOException, InterruptedException {
        ProcessBuilder pb = new ProcessBuilder(cmds);
        pb = pb.redirectErrorStream(true);
        pb = pb.directory(dir);

        if (!AppUtilities.isWindows()) {
            //	Add more directories to the search path where commands might be found.
            //	Java apps on MacOS have only the default system path and not the user path,
            //	so this is necessary on that platform and harmless on other Unix type systems.
            Map<String, String> env = pb.environment();
            String path = env.get("PATH");
            env.put("PATH", path + ":/usr/local/bin:/opt/local/bin:/sw/bin:/opt/homebrew/bin:~/bin:.");
        }

        Process proc = pb.start();			//	Start the process.

        //	Collect the output from the process.
        StreamGobbler stdout = new StreamGobbler(proc.getInputStream(), "OUTPUT");
        stdout.start();
        if (wait)
            proc.waitFor();
        else
            Thread.sleep(1000 * 2);	//	Sleep this thread for a couple of seconds.

        //  Collect the output as a list and return it.
        stdout.join(3000);
        stdout.close();
        proc.destroy();
        List<String> messages = stdout.getLines();

        return messages;
    }

    /**
     * Method that, on UNIX platforms ONLY, uses the "which" command to locate the
     * specified executable on the search path.
     *
     * @param name The name of the executable to search for.
     * @param defaultPath The default path to return if the executable can't be found.
     * @return The path to the executable or the defaultPath if the executable can't be
     * found.
     */
    private static String findExecutablePath(String name, String defaultPath) {
        //  Assumes we have a UNIX-like system.
        List<String> cmds = new ArrayList();
        cmds.add("which");
        cmds.add(name);
        File dir = new File(System.getProperty("user.dir"));
        try {
            List<String> output = runCommand(cmds, dir, false);
            if (!output.isEmpty()) {
                String line = output.get(0);
                if (!line.contains("No such file")) {
                    return line;
                }
            }
        } catch (Exception ignore) { }
        
        return defaultPath;
    }

    /**
     * Returns true if this system has the required programs on the search path.
     */
    public boolean hasODIE() {

        try {
            File dir = new File(System.getProperty("user.dir"));

            //  Look for ODIE
            List<String> cmds = new ArrayList();
            cmds.add(getODIEPath());
            List<String> output = runCommand(cmds, dir, false);
            if (output.size() < 3)
                return false;
            String line = output.get(2);
            if (!line.contains("Odie")) {
                return false;
            }

        } catch (Exception e) {
            return false;
        }

        return true;
    }
}
