/*
 *  ODIEInterface -- A class that handles the complexity of interacting with ODIE.
 *
 *  Copyright (C) 2013-2017, Joseph A. Huwaldt. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
package jahuwaldt.jodie;

import jahuwaldt.io.StreamGobbler;
import jahuwaldt.swing.AppUtilities;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * Provides an interface for interacting with ODIE. This interface instantiates an
 * instance of ODIE when it is created. Then the user may pass commands to ODIE using
 * "cmds()" and get output back from ODIE using "output()". When the user wants the
 * instance of ODIE to be closed, they should call the "close()" method.
 *
 * @author Joseph A. Huwaldt Date: November 5, 2013
 * @version March 13, 2017
 */
public class ODIEInterface {

    private static final int MAXWAIT = 500;
    
    //  A reference to the ODIE process.
    private final Process _p;
    
    //  Flags indicating if input and output should be echoed to stdout.
    private boolean _echoInput = false;
    private boolean _echoOutput = false;
    
    //  A writer for sending commands to ODIE.
    private final PrintWriter _pwrite;
    
    //  A reader for getting output from ODIE.
    private final StreamGobbler _pread;
    

    /**
     * Initializes an instance of an ODIEInterface object.
     *
     * @param odieExe A reference to the ODIE executable file. If <code>null</code> then
     *       "odie" will be used.
     * @param workingDir The working directory for the process. May pass <code>null</code>
     *      if a working directory does not need to be specified.
     */
    @SuppressWarnings("CallToThreadStartDuringObjectConstruction")
    public ODIEInterface(File odieExe, File workingDir) throws IOException {

        String odieCmd = "odie";
        if (odieExe != null && odieExe.exists() && odieExe.isFile())
            odieCmd = odieExe.getAbsolutePath();

        String[] cmds = {"/bin/sh", "-c", odieCmd, ">/dev/tty"};
        if (AppUtilities.isWindows()) {
            cmds[0] = "CMD.EXE";
            cmds[1] = "/C";
            cmds[3] = "CON:";
        }

        ProcessBuilder pb = new ProcessBuilder(cmds);
        pb = pb.redirectErrorStream(true);
        if (workingDir != null && workingDir.isDirectory())
            pb = pb.directory(workingDir);
        _p = pb.start();			//	Start the process.

        //  Connect a print writer to stdin.
        _pwrite = new PrintWriter(_p.getOutputStream());

        //  Connect a reader to stdout.
        _pread = new StreamGobbler(_p.getInputStream(), "OUTPUT");
        _pread.start();
    }

    public void setEchoInput(boolean echo) {
        _echoInput = echo;
    }

    public boolean echoInput() {
        return _echoInput;
    }

    public void setEchoOutput(boolean echo) {
        _echoOutput = echo;
    }

    public boolean echoOutput() {
        return _echoOutput;
    }

    /**
     * Method that waits up to 5 seconds until the specified character string appears in
     * the ODIE output.
     *
     * @param key The key to look for in the ODIE output.
     * @return All the lines of output preceding and including the specified key.
     * @throws InterruptedException
     */
    public List<String> waitForKey(String key) throws InterruptedException {
        List<String> output = new ArrayList();
        List<String> recent = _pread.getLinesAndClear();
        output.addAll(recent);
        if (_echoOutput)
            for (String line : recent) {
                System.out.println(line);
            }
        int count = 0;
        while (count < MAXWAIT && (recent.isEmpty() || !recent.get(recent.size() - 1).contains(key))) {
            ++count;
            Thread.sleep(100);
            recent = _pread.getLinesAndClear();
            output.addAll(recent);
            if (_echoOutput)
                for (String line : recent) {
                    System.out.println(line);
                }
        }
        return output;
    }

    /**
     * Method that waits up to 5 seconds until the specified number of lines of output
     * have been returned from ODIE.
     *
     * @param numLines The number of lines of output from ODIE to wait for.
     * @return The specified number of lines of output from ODIE.
     * @throws InterruptedException
     */
    public List<String> waitForLine(int numLines) throws InterruptedException {
        int count = 0;
        while (count < MAXWAIT && _pread.getNumberOfLines() < numLines) {
            Thread.sleep(100);
        }
        return _pread.getLinesAndClear();
    }

    /**
     * Sends a command to ODIE.
     *
     * @param cmd The command to send.
     */
    public void cmd(String cmd) {
        if (_echoInput)
            System.out.println(cmd);
        _pwrite.println(cmd);
        _pwrite.flush();
    }

    /**
     * Closes ODIE. After calling this method, this class instance is no longer
     * functional.
     */
    public void close() throws IOException {
        //  Tell ODIE to quit.
        cmd("Quit");
        try {
            Thread.sleep(10);
        } catch (InterruptedException ignore) { }

        //  Close the input and output streams.
        _pwrite.close();
        _pread.close();

        //  Kill the ODIE process (if it hasn't already quit).
        _p.destroy();
    }
}
