/*
 *  RunODIEThread -- A thread that runs the ODIE program and gather's it's output.
 *
 *  Copyright (C) 2013-2024, Joseph A. Huwaldt. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
package jahuwaldt.jodie;

import jahuwaldt.io.FileUtils;
import jahuwaldt.js.tables.FTableList;
import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.js.tables.OTISTableReader;
import jahuwaldt.js.util.TextTokenizer;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.ProgressBarHandler;
import jahuwaldt.tools.math.MathTools;
import java.awt.Component;
import java.io.*;
import java.text.MessageFormat;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A thread used to run the ODIE program and plot up the results.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt,    Date: November 4, 2013
 * @version January 19, 2024
 */
public class RunODIEThread extends Thread {
    
	//  Debug flag.
	private static final boolean DEBUG = false;
    
    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(RunODIEThread.class.getName());

    //  The name of the OTIS table file passed to ODIE.
    private static final String TBLFILE = "orig_tables.dat";
    
    //  The name of the ODIE plot file output from ODIE.
    private static final String PLOTFILE = "plot.txt";
    
    //  The maximum number of interpolated breakpoints to allow.
    private static final int MAXINTERPBP = 500000;
    
    //  The amount that ODIE should enrich each set of breakpoints
    private static final int KENR = 10;
	
	//	The parent component for this thread.
	private final Component parent;
	
	//	A list of buttons to be disabled while POST is running.
	private final List<JButton> disableLst;
	
	//	A list of buttons to be enabled while POST is running.
	private final List<JButton> enableLst;
	
    //  The tables being plotted.
    private final FTableList tables;
    
	//  The path to the ODIE executable on the local system.
	private final File odieExe;
	
	//	The process for the running instance of ODIE.
	private ODIEInterface odie;
	
    
	/**
	*  Construct an thread that can run the POST3D program.
	*
	*  @param  parent      The parent component for this thread's dialogs.
	*  @param  inpTables   A list of tables to be processed by ODIE.
	*  @param  disable     A list of buttons to disable while ODIE is running.  Pass <code>null</code>
    *                       if there are no buttons to disable.
	*  @param  enable      A list of buttons to enable while ODIE is running.  Pass <code>null</code>
    *                       if there are no buttons to enable.
	*  @throws IOException if any of the inputs are invalid.
	*/
	public RunODIEThread(Component parent, FTableList inpTables, List<JButton> disable, List<JButton> enable) throws IOException {
        this.parent = parent;
        this.tables = inpTables;
		this.disableLst = disable;
		this.enableLst = enable;
        MDIApplication guiApp = MDIApplication.getInstance();
        ResourceBundle resBundle = guiApp.getResourceBundle();
        
		String odiePath = guiApp.getPreferences().get(AppPreferences.ODIEPATH);
		if (odiePath == null) {
			guiApp.getPreferences().showPreferenceDialog();
			throw new IOException(resBundle.getString("undefinedODIEPath"));
		}
		this.odieExe = new File(odiePath);
		if (!odieExe.exists()) {
			guiApp.getPreferences().showPreferenceDialog();
			throw new IOException(resBundle.getString("invalidODIEPath"));
		}
		
    }
    
	/**
	*  This method is called automatically by the Thread class
	*  and should not be called by any other classes.  Call the
	*  Thread method "start()", to start this thread running.
	*/
    @Override
	public void run() {
		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
		
		if (DEBUG)	System.out.println("Running...");
		
        ProgressBarHandler progBar = new ProgressBarHandler(resBundle.getString("progBarLabel"), null);
        
        File tmpDir = null;
        try {
            //  Create a temporary directory to run ODIE in.
            tmpDir = FileUtils.createTempDirectory("odie");

            //  Write an OTIS format table file containing the tables to be processed
            //  into the temporary directory.
            File tableFile = new File(tmpDir, TBLFILE);
            OTISTableReader writer = new OTISTableReader();
            OutputStream os = new FileOutputStream(tableFile);
            writer.write( os, tables );
            os.close();
            
            if (DEBUG)
                System.out.println("tableFile = " + tableFile.getAbsolutePath());

            //	Adjust the button states as required (on the Event Dispatch Thread only!).
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    if (disableLst != null)
                        for (JButton btn : disableLst) {
                            btn.setEnabled(false);
                        }
                    if (enableLst != null)
                        for (JButton btn : enableLst) {
                            btn.setEnabled(true);
                        }
                }
            });

            //  Start up the ODIE process.
            odie = new ODIEInterface(odieExe, tmpDir);
            if (DEBUG) {
                odie.setEchoInput(true);
                odie.setEchoOutput(true);
            }
            
             //  Wait for ODIE to ask for input.
            odie.waitForKey("?");
            
            //  Tell ODIE to load the table file we just saved out above.
            progBar.setNote(resBundle.getString("progBarNoteReadingTbls"));
            progBar.setProgress(0.1F);
            if (progBar.isCanceled()) return;
            odie.cmd(TBLFILE);
            odie.waitForKey("?");
            
            
            //  Loop over all the tables to be analyzed.
            float progRange = 0.9F/tables.size();
            for (Float64Table tbl : tables) {
                if (progBar.isCanceled()) return;
                progBar.setNote(
                        MessageFormat.format(resBundle.getString("progBarNoteDetBP"),
                        tbl.getName()));
                
                //  Tell ODIE to work with this table.
                String name = OTISTableReader.sanitizeTableName(tbl.getName()).toString();
                odie.cmd(name);
                odie.waitForKey("?");
                
                //  Tell ODIE that we want a plot file created.
                odie.cmd("Y");
                
                // The number of breakpoints for each dimension of the plot file.
                FastTable<Integer> indepNum = FastTable.newInstance();
                
                //  Come up with a reasonable number of breakpoints to insert.
                int numDims = tbl.getDimensions();
                int nEnr = KENR;
                int size = tbl.size();
                if (size * MathTools.pow(nEnr, numDims) > MAXINTERPBP) {
                    double exp = Math.log10(MAXINTERPBP / ((double) size)) / numDims;
                    nEnr = (int) Math.pow(10, exp);
                    if (nEnr < 3)
                        nEnr = 3;
                    if (DEBUG)
                        System.out.println("nEnr = " + nEnr);
                }

                //  Loop over each dimension asking ODIE to interpolate each.
                for (int dim=0; dim < numDims; ++dim) {
                    String indepName = OTISTableReader.sanitizeIndepName(tbl.getIndepName(dim)).toString();
                    int nOrigBP = tbl.getNumBreakpoints(dim);
                    
                    if (dim < numDims - 1) {
                        odie.waitForKey("?");
                        odie.cmd(indepName);
                    }
                    
                    //  Tell ODIE to ask for a list of breakpoints in G10.7 format.
                    odie.waitForKey(indepName);
                    odie.cmd("2");
                    
                    //  Tell ODIE the set of breakpoints to use in this dimension.
                    odie.waitForKey(indepName);
                    StringBuilder buffer = new StringBuilder();
                    int nNewBP = 0;
                    for (int bpIdx = 0; bpIdx < nOrigBP-1; ++bpIdx) {
                        double thisBP = tbl.getBreakpointValue(dim,bpIdx);
                        buffer.append(formatNumberG10p7(thisBP));
                        ++nNewBP;
                        double nextBP = tbl.getBreakpointValue(dim,bpIdx+1);
                        double dx = (nextBP - thisBP)/nEnr;
                        for (int i=1; i < nEnr; ++i) {
                            double newBP = thisBP + dx*i;
                            buffer.append(formatNumberG10p7(newBP));
                            ++nNewBP;
                       }
                    }
                    buffer.append(formatNumberG10p7(tbl.getBreakpointValue(dim,nOrigBP-1)));
                    ++nNewBP;
                    odie.cmd(buffer.toString());
                    
                    //  Store the number of breakpoints in this dimension.
                    indepNum.add(nNewBP);
                    
                    //  Update progress bar.
                    if (progBar.isCanceled()) return;
                    progBar.incrementProgress(progRange*0.1F/numDims);
                }
                
                //  Tell ODIE the name of the plot file we want to use.
                if (progBar.isCanceled()) return;
                progBar.setNote(
                        MessageFormat.format(resBundle.getString("progBarNoteRunODIE"),
                        tbl.getName()));
                odie.waitForKey("Name");
                odie.cmd(PLOTFILE);
                odie.waitForKey("?");
                if (progBar.isCanceled()) return;
                progBar.incrementProgress(progRange*0.8F);
                
                //  Process the plot file.
                Float64Table interpTbl = readPlotFile(tbl, tmpDir, indepNum);
 
                //  Create a plot window for this table and it's interpolation (on the Event Dispatch Thread).
                SwingUtilities.invokeLater(new PlotTableThread(tbl, interpTbl));
                
                if (progBar.isCanceled()) return;
                progBar.incrementProgress(progRange*0.1F);
                progBar.setNote(MessageFormat.format(resBundle.getString("progBarNoteDone"),
                        tbl.getName()));
                
            }   //  Next tbl
            
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, null, e);
			
			if (parent != null) {
				String msg = e.getMessage();
				if (msg == null) {
					Throwable cause = e.getCause();
					if (cause != null)
						msg = cause.getMessage();
                }
                final String msg2 = msg;
                
                //  Can only call showMessageDialog() on the Event Dispatch Thread.
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        JOptionPane.showMessageDialog(parent, msg2,
                                MDIApplication.getInstance().getResourceBundle().getString("ioErrorTitle"),
                                JOptionPane.ERROR_MESSAGE);
                    }
                });
            }
		
        } catch (final ArithmeticException e) {
			LOGGER.log(Level.SEVERE, null, e);
			
			if (parent != null) {
                //  Can only call showException() on the Event Dispatch Thread.
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                        AppUtilities.showException(parent, resBundle.getString("unexpectedTitle"),
								resBundle.getString("unexpectedMsg"), e);
                    }
                });
            }
            
        } catch (final InterruptedException e) {
            LOGGER.log(Level.SEVERE, null, e);
            
            if (parent != null) {
                //  Can only call showException() on the Event Dispatch Thread.
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                        AppUtilities.showException(parent, resBundle.getString("unexpectedTitle"),
                                resBundle.getString("unexpectedMsg"), e);
                    }
                });
            }
        } finally {
            //  Make sure to kill the ODIE process before leaving.
            if (odie != null) {
                try {
                    odie.close();
                } catch (IOException e) {
                    LOGGER.log(Level.SEVERE, null, e);
                }
                odie = null;
            }

            //  Delete the temporary files before leaving.
            if (tmpDir != null) {
                FileUtils.deleteDirectory(tmpDir);
            }

            progBar.setProgress(1);
            
            //	Adjust the button states as required (on the Event Dispatch Thread only!).
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    if (disableLst != null)
                        for (JButton btn : disableLst) {
                            btn.setEnabled(true);
                        }
                    if (enableLst != null)
                        for (JButton btn : enableLst) {
                            btn.setEnabled(false);
                        }
                }
            });
        }
        
		if (DEBUG)	System.out.println("Done");
    }
    
    /**
     * Method that formats a double in FORTRAN G10.7 format.
     * @param value The number to be formatted.
     * @return A string representation of the number in G10.7 format.
     */
    private static String formatNumberG10p7(double value) {
        String numStr = String.format("%10.7g", value);
        if (numStr.length() > 10)
            numStr = numStr.substring(0,10);
        return numStr;
    }

    /**
     * A runnable that creates a plot window on the Event Dispatch Thread
     * when combined with SwingUtilities.invokeLater().
     */
    private class PlotTableThread implements Runnable {
        private final Float64Table _tbl;
        private final Float64Table _interpTbl;
        
        public PlotTableThread(Float64Table tbl, Float64Table interpTbl) {
            _tbl = tbl;
            _interpTbl = interpTbl;
        }
 
        @Override
        public void run() {
            try {
                
                //  Create a plot window for this table and it's interpolation.
                String title = MessageFormat.format(
                        MDIApplication.getInstance().getResourceBundle().getString("plotText"),
                        _tbl.getName());
                TablePlotWindow plot = new TablePlotWindow(title, _tbl, _interpTbl);
                plot.setVisible(true);
                
            } catch (NoSuchMethodException e) {
                LOGGER.log(Level.SEVERE, null, e);
                if (parent != null) {
                    ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                    AppUtilities.showException(parent, resBundle.getString("unexpectedTitle"),
                            resBundle.getString("unexpectedMsg"), e);
                }
            }

        }
    }
    
    /**
     * Reads in the plot file from ODIE and returns the table of interpolated values.
     * 
     * @param tbl   The original un-interpolated table.
     * @param dir   The directory containing the plot file.
     * @param indepNum  The number of breakpoints in each dimension of the interpolated table.
     * @return A table of interpolated points from ODIE.
     * @throws IOException 
     */
    private Float64Table readPlotFile(Float64Table tbl, File dir, List<Integer> indepNum) throws IOException {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        
		// Set up the tokenizer options.
		TextTokenizer tokenizer = TextTokenizer.newInstance();
		tokenizer.setDelimiters(" ");

        //  Get some info for the interpolated table.
        List<Text> indepNames = tbl.getIndepNames();
        String tblName = tbl.getName() + "_interp";
        int numDims = tbl.getDimensions();
        
        // Create empty breakpoints correctly sized for each dimension.
		int numElements = 1;
        FastTable<Float64Vector> breakpoints = FastTable.newInstance();
		for (int i = 0; i < numDims; ++i ) {
			int num = indepNum.get(i);
			FastTable<Float64> row = FastTable.newInstance();
			for (int j=0; j < num; ++j)
				row.add(Float64.ZERO);
			breakpoints.add(Float64Vector.valueOf(row));
			FastTable.recycle(row);
			numElements *= num;
		}
        
        //  Create an empty table of the correct dimensions.
        Float64Table interpTbl = Float64Table.valueOf( tblName, indepNames, breakpoints, null );
        
        // Set up the position index table and a table for tracking breakpoints.
        FastTable<Integer64> pos = FastTable.newInstance();
		FastTable<Float64> old = FastTable.newInstance();
        for (int i=0; i < numDims; ++i) {
            pos.add(Integer64.ZERO);
			old.add(Float64.ZERO);
        }
        
        //  Get an input stream reader to the plot file.
        File plotFile = new File(dir, PLOTFILE);
        try (FileInputStream fis = new FileInputStream(plotFile)) {
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(fis));

            //  Skip the header line.
            reader.readLine();

            for (int i = 0; i < numElements; ++i) {
                // Read in the next line from the file.
                String aLine = reader.readLine();
                if (aLine == null)
                    throw new IOException(
                            MessageFormat.format(resBundle.getString("eofErrMsg"),
                            reader.getLineNumber()));
                if (aLine.trim().equals("")) {
                    aLine = reader.readLine();
                    if (aLine == null)
                        throw new IOException(
                                MessageFormat.format(resBundle.getString("eofErrMsg"),
                                reader.getLineNumber()));
                }
                tokenizer.setText(aLine);

                //	Read in the independent values.
                for (int dim = 0; dim < numDims; ++dim) {
                    try {

                        // Read in the independent values.
                        Float64 value = Float64.valueOf(tokenizer.nextToken().trim());
                        if (i == 0) {
                            // First time through, store off old values.
                            old.set(dim, value);

                        } else {
                            // Only increment this dimension's index if the value has changed.
                            if (!value.equals(old.get(dim))) {
                                old.set(dim, value);
                                pos.set(dim, pos.get(dim).plus(Integer64.ONE));		//	++pos[dim]
                                // Once we've reached the max breakpoint, go back to the 1st.
                                if (pos.get(dim).intValue() >= breakpoints.get(dim).getDimension())
                                    pos.set(dim, Integer64.ZERO);
                            }
                        }

                        //	Change the breakpoint value in all the tables.
                        interpTbl.setBreakpoint(dim, pos.get(dim).intValue(), value);

                    } catch (NoSuchElementException e) {
                        LOGGER.log(Level.SEVERE, null, e);
                        throw new IOException(
                                MessageFormat.format(resBundle.getString("valueNotFoundMsg"),
                                indepNames.get(dim), reader.getLineNumber()));
                    } catch (NumberFormatException e) {
                        LOGGER.log(Level.SEVERE, null, e);
                        throw new IOException(
                                MessageFormat.format(resBundle.getString("couldNotParseMsg"),
                                indepNames.get(dim), reader.getLineNumber()));
                    }
                }   //  Next dim

                // Read in the dependent value.
                try {
                    Text token = tokenizer.nextToken();
                    //System.out.println("token = " + token);
                    Float64 value = Float64.valueOf(token.trim());
                    interpTbl.set(pos, value);

                } catch (NoSuchElementException e) {
                    throw new IOException(
                            MessageFormat.format(resBundle.getString("valueNotFoundMsg"),
                            tbl.getName(), reader.getLineNumber()));
                } catch (NumberFormatException e) {
                    throw new IOException(
                            MessageFormat.format(resBundle.getString("couldNotParseMsg"),
                            tbl.getName(), reader.getLineNumber()));
                } catch (Exception e) {
                    throw new IOException(
                            MessageFormat.format(resBundle.getString("couldNotParseMsg"),
                            tbl.getName(), reader.getLineNumber()));
                }

            }   //  Next i

        } finally {
            //  Clean up before leaving.
            FastTable.recycle(breakpoints);
            FastTable.recycle(pos);
            FastTable.recycle(old);
            TextTokenizer.recycle(tokenizer);
        }
        
        return interpTbl;
    }
}
