/*
*  SelectPlotDimsDialog  -- Dialog allowing user to select a subset of a list of plot dimension variables.
*
*  Copyright (C) 2008-2024, Joseph A. Huwaldt. All rights reserved.
*
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
package jahuwaldt.jodie;

import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.swing.EscapeJDialog;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Frame;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;
import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

/**
 * A modal dialog the allows the user to select a subset of all the available dimensions
 * in a table.
 *
 * <p> Modified by: Joseph A. Huwaldt    </p>
 *
 * @author Joseph A. Huwaldt, Date: February 4, 2008
 * @version January 19, 2024
 */
public class SelectPlotDimsDialog extends EscapeJDialog implements ActionListener {

    //	The column name at the top of the available dimensions column.
    private static final String AVAIL_COLNAME = "Available Dimensions";

    //	The column name at the top of the selected dimensions column.
    private static final String SEL_COLNAME = "Selected Dimensions ";

    //	A list of strings that show up as the rows in the available dimensions table.
    private ArrayList<String> availRowData = new ArrayList();

    /**
     * The list of available dimensions (represented by Integers) in the table.
     */
    private List<Integer> availElements = new ArrayList();

    /**
     * The JTable displaying the contents of the "availElements" list that is available.
     */
    private JTable availTable;

    /**
     * The JTable displaying the elements that the user has selected. This table is only
     * used if multiple selections are possible.
     */
    private final JTable selectedTable;

    //	A list of the dimensions (represented by Integers) that were selected by the user.
    private List<Integer> selectedDims;

    //	A list of lists of selected elements.
    private List selRowData = new ArrayList();
    private List<Integer> selElements = new ArrayList();

    /**
     * Construct a model dialog that allows the user to select a subset of all the
     * available dimensions in a table.
     *
     * @param parent   A reference to the parent frame that this dialog belongs to.
     * @param title    The title for the dialog window.
     * @param message  A message to display telling the user what to do.
     * @param data     A multi-dimensioned table.
     * @param dimOrder The order of the dimensions in the available table.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public SelectPlotDimsDialog(Frame parent, String title, String message, Float64Table data, int[] dimOrder) {
        super(parent, title, true);

        //	Copy all the dimensions from the table into the table lists.
        int numDims = data.getDimensions();
        for (int i = numDims - 1; i >= 0; --i) {
            availElements.add(dimOrder[i]);
            String element = data.getIndepName(dimOrder[i]).toString();
            availRowData.add(element);
        }

        //  Layout the dialog window.
        Container cp = this.getContentPane();
        cp.setLayout(new BorderLayout());

        //	Add instructions at the top.
        Box topPanel = Box.createHorizontalBox();
        cp.add(topPanel, BorderLayout.NORTH);

        topPanel.add(Box.createHorizontalStrut(10));
        JLabel label = new JLabel(message);
        label.setBorder(BorderFactory.createEmptyBorder(6, 6, 6, 6));
        topPanel.add(label);
        topPanel.add(Box.createGlue());

        //	Create a panel to display in the center.
        Box centerPanel = Box.createHorizontalBox();
        cp.add(centerPanel, BorderLayout.CENTER);

        //  Create a list that displays the cases in all the input data sets.
        availTable = new JTable(new AbstractTableModel() {
            @Override
            public String getColumnName(int col) {
                return SelectPlotDimsDialog.this.getAvailColumnName(col);
            }

            @Override
            public int getRowCount() {
                return availRowData.size();
            }

            @Override
            public int getColumnCount() {
                return 1;
            }

            @Override
            public Object getValueAt(int row, int col) {
                return availRowData.get(row);
            }

            @Override
            public boolean isCellEditable(int row, int col) {
                return false;
            }

            @Override
            public void setValueAt(Object value, int row, int col) {
            }
        });
        availTable.setCellSelectionEnabled(false);
        availTable.setColumnSelectionAllowed(false);
        availTable.setRowSelectionAllowed(true);
        availTable.clearSelection();
        availTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);	//	Show hoirz. scroll bars.
        availTable.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    doAdd();
                }
            }
        });
        availTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        //	If only a single row exists, preselect it.
        if (availRowData.size() == 1)
            availTable.selectAll();

        //	Initialize the column widths to show all the data.
        initColumnWidths(availTable);

        //  Put the list in a scroll pane so that we have scroll bars.
        JScrollPane scrollPane = new JScrollPane(availTable);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

        //  Add the list to the panel.
        centerPanel.add(scrollPane);

        //	Create a button panel for selection buttons.
        centerPanel.add(Box.createHorizontalStrut(10));
        centerPanel.add(createButtonPanel());

        //	Create an output table that displays selected tables.
        selectedTable = new JTable(new AbstractTableModel() {
            @Override
            public String getColumnName(int col) {
                return SelectPlotDimsDialog.this.getSelColumnName(col);
            }

            @Override
            public int getRowCount() {
                return selRowData.size();
            }

            @Override
            public int getColumnCount() {
                return 1;
            }

            @Override
            public Object getValueAt(int row, int col) {
                return selRowData.get(row);
            }

            @Override
            public boolean isCellEditable(int row, int col) {
                return false;
            }

            @Override
            public void setValueAt(Object value, int row, int col) {
            }
        });
        selectedTable.setCellSelectionEnabled(false);
        selectedTable.setColumnSelectionAllowed(false);
        selectedTable.setRowSelectionAllowed(true);
        selectedTable.clearSelection();
        selectedTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);	//	Show hoirz. scroll bars.
        selectedTable.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    doRemove();
                }
            }
        });
        selectedTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        //	Initialize the column widths to show all the data.
        initColumnWidths(selectedTable);

        //  Put the list in a scroll pane so that we have scroll bars.
        scrollPane = new JScrollPane(selectedTable);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

        //  Add the list to the panel.
        centerPanel.add(Box.createHorizontalStrut(10));
        centerPanel.add(scrollPane);

        //	Set a default window size.
        setSize(450, 300);

        //  Define OK and Cancel buttons.
        Box box = Box.createHorizontalBox();
        box.add(Box.createGlue());

        JButton cancelBtn = new JButton("Cancel");
        cancelBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                performEscapeAction(null);
            }
        });
        box.add(cancelBtn);

        box.add(Box.createHorizontalStrut(40));

        //	If multiple selections allowed, change "OK" button text to "Done".
        String okMessage = "Done";
        JButton okayBtn = new JButton(okMessage);
        okayBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                SelectPlotDimsDialog.this.actionPerformed(null);
            }
        });
        this.getRootPane().setDefaultButton(okayBtn);
        box.add(okayBtn);

        box.add(Box.createHorizontalStrut(20));
        cp.add(box, BorderLayout.SOUTH);

    }

    /**
     * This method returns a list object that contains Integer objects representing the
     * table dimension that were selected in this dialog. If the user has not clicked on
     * the "OK" button, null is returned.
     */
    public List<Integer> getSelected() {
        return selectedDims;
    }

    /**
     * Method that creates a button panel containing selection buttons.
     */
    private Box createButtonPanel() {

        //	Create selection buttons to run down the center.
        Box btnBox = Box.createVerticalBox();
        btnBox.add(Box.createGlue());

        //	Create an "Add" button.
        JButton button = new JButton("--->");
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doAdd();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createVerticalStrut(10));

        //	Create a "Remove" button.
        button = new JButton("<---");
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRemove();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createVerticalStrut(10));

        //	Create an "Add All" button.
        button = new JButton("Add All");
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doAddAll();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createVerticalStrut(10));

        //	Create a "Remove All" button.
        button = new JButton("Remove All");
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRemoveAll();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createGlue());

        return btnBox;
    }

    /**
     * This method is called when the user clicks on the "Add" button.
     */
    private void doAdd() {

        //	Copy the selected elements to the output table.
        int[] selected = availTable.getSelectedRows();
        int numSelected = selected.length;

        for (int i = 0; i < numSelected; ++i) {
            Integer element = availElements.get(selected[i]);
            selElements.add(element);
            selRowData.add(availTable.getValueAt(selected[i], 0));
            availElements.remove(selected[i]);
            availRowData.remove(selected[i]);
        }

        //	Update the output table's UI.
        selectedTable.revalidate();
        availTable.revalidate();

    }

    /**
     * This method is called when the user clicks on the "Add All" button.
     */
    private void doAddAll() {

        int numRows = availTable.getRowCount();
        for (int i = 0; i < numRows; ++i) {
            Integer element = availElements.get(i);
            selElements.add(element);
            selRowData.add(availTable.getValueAt(i, 0));
        }

        availElements.clear();
        availRowData.clear();

        selectedTable.revalidate();
        availTable.revalidate();

    }

    /**
     * This method is called when the user clicks on the "Remove" button.
     */
    private void doRemove() {

        //	Get a list of all the selected rows in the table.
        int[] selected = selectedTable.getSelectedRows();
        int numSelected = selected.length;

        for (int i = 0; i < numSelected; ++i) {
            Integer element = selElements.get(i);
            availElements.add(element);
            availRowData.add((String) selectedTable.getValueAt(i, 0));
            selElements.remove(selected[i]);
            selRowData.remove(selected[i]);
        }

        selectedTable.revalidate();
        availTable.revalidate();
    }

    /**
     * This method is called when the user clicks on the "Remove All" button.
     */
    private void doRemoveAll() {

        int numRows = selectedTable.getRowCount();
        for (int i = 0; i < numRows; ++i) {
            Integer element = selElements.get(i);
            availElements.add(element);
            availRowData.add((String) selectedTable.getValueAt(i, 0));
        }

        selElements.clear();
        selRowData.clear();

        selectedTable.revalidate();
        availTable.revalidate();
    }

    /**
     * Method that returns a list of selected elements.
     */
    protected Integer[] getSelectedElements() {

        Integer[] selList;

        //	Return a list of all the elements in the output table.
        int totRows = selectedTable.getRowCount() + availTable.getRowCount();
        selList = new Integer[totRows];
        int selRows = selectedTable.getRowCount();
        int pos = 0;
        for (int i = 0; i < selRows; ++i, ++pos)
            selList[i] = selElements.get(i);

        for (int i = pos; i < totRows; ++i)
            selList[i] = availElements.get(i - selRows);

        //	Reverse the list.
        int j = totRows - 1;
        int num = totRows / 2;
        for (int i = 0; i < num; ++i, --j) {
            Integer tmp = selList[i];
            selList[i] = selList[j];
            selList[j] = tmp;
        }

        return selList;
    }

    /**
     * This method is called when the user clicks on the "OK Button".
     */
    @Override
    public void actionPerformed(ActionEvent evt) {

        //	Get a list of all the selected rows in the table.
        Integer[] selected = getSelectedElements();
        int numSelected = selected.length;

        //	Add selected cases to the selected data set.
        selectedDims = new ArrayList<>();
        for (int i = 0; i < numSelected; ++i)
            selectedDims.add(selected[i]);

        //  Hide the window.
        this.setVisible(false);
    }

    /**
     * Response to ENTER key pressed goes here. This has the same result as clicking on
     * the OK or Select button.
     */
    @Override
    protected void performEnterAction(KeyEvent e) {
        actionPerformed(null);
    }

    /**
     * Method that returns the name to place above the available dimensions column.
     */
    protected String getAvailColumnName(int col) {
        return AVAIL_COLNAME;
    }

    /**
     * Method that returns the name to place above the selected dimensions column.
     */
    protected String getSelColumnName(int col) {
        return SEL_COLNAME;
    }

    /**
     * Method that resets the minimum column widths.
     */
    private static void initColumnWidths(JTable table) {
        int numColumns = table.getColumnCount();
        for (int i = 0; i < numColumns; ++i) {
            //	Get each column.
            TableColumn column = table.getColumnModel().getColumn(i);

            //	Determine the header width.
            int headerWidth = 0;
            try {
                TableCellRenderer renderer = column.getHeaderRenderer();
                if (renderer == null)
                    renderer = table.getTableHeader().getDefaultRenderer();

                Component comp = renderer.
                        getTableCellRendererComponent(
                                null, column.getHeaderValue(),
                                false, false, 0, 0);
                headerWidth = comp.getPreferredSize().width;
            } catch (NullPointerException e) {
                e.printStackTrace();
            }

            //	Set the minimum width to the larger of the header width
            //	and 150 pixels.
            column.setMinWidth(Math.max(50, headerWidth) + 10);
        }
    }

}
