/*
 *  TablePlotWindow -- Window that plots selected n-dimensional tables.
 *
 *  Copyright (C) 2013-2024, Joseph A. Huwaldt. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
package jahuwaldt.jodie;

import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.DecimalField;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.util.GeneralFormat;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.List;
import java.util.ResourceBundle;
import javax.swing.*;
import javolution.util.FastTable;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * An application window for the application jODIE that displays a plot of a
 * selected n-dimensional table.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: November 5, 2013
 * @version January 19, 2024
 */
public class TablePlotWindow extends AbstractPlotWindow {
    //  Numberical tolerance.
    private static final double EPS = 1e-15;
    
	//	The content pane for our window.
	private final JPanel _centerPanel = new JPanel(new BorderLayout());
	
	//	References to the left/right buttons.
	private final JButton _leftButton;
	private final JButton _rightButton;

	// The un-interpolated table we are doing the plot of.
	private final Float64Table _origTbl;
    
    //  The ODIE interpolated table we are going to overlay.
    private final Float64Table _interpTbl;

	// A reference to the panel containing the current high dimension breakpoints.
	private JPanel _bpPanel;
	
	// The number format used to format numbers.
	private final DecimalFormat _nf = (DecimalFormat)NumberFormat.getInstance();
	
	//	Storage for indexes into multi-dimensioned tables.
	private final List<Integer64> _index;
	
	//	Indicies that give the order of the plotting of dimensions in the table.
	private final int[] _dimOrder;
	
    //  A text field for inputting the number that is to be represented as undefined.
    private final DecimalField _undefinedTF;
    

	/**
	*  Creates a new instance of a table plot window using the
	*  specified table and it's interpolation.
	*
	* @param origTable      The original un-interpolated table we want to do a plot of.
    * @param interpTable    The interpolated table out of ODIE.  Must be the same dimensions, with the same
    *                       independent variables as origTable, but with a greater number of breakpoints
    *                       in each dimension.
	*/
    @SuppressWarnings("OverridableMethodCallInConstructor")
	public TablePlotWindow( String title, Float64Table origTable, Float64Table interpTable ) throws NoSuchMethodException {
		super( title, origTable.getName().toString() );

		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        
		setResizable( true );
		setSize(1020, 640);
		
		//  Position the window so that each new window can be seen.
		AppUtilities.positionWindow( this, getWidth(), getHeight() );

		//  Have the window dispose of itself when it closes.
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		// Set up some class variables.
		_origTbl = origTable;
        _interpTbl = interpTable;
		_nf.setMinimumFractionDigits(3);
		
		int numDim = origTable.getDimensions();
		_index = FastTable.newInstance();
		for (int i=0; i < numDim; ++i)
			_index.add(Integer64.ZERO);
		
		_dimOrder = new int[numDim];
		for (int i=0; i < numDim; ++i)
			_dimOrder[i] = i;

		// Set up this window's user interface.
		getContentPane().add(_centerPanel);
		
        //  Create an "undefined" value entry field.
        JLabel undefinedLabel = new JLabel(resBundle.getString("undefinedLabel") + " ");
        _undefinedTF = new DecimalField(Double.NaN, 8, new GeneralFormat());
        _undefinedTF.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed( ActionEvent e ) {
                regeneratePlot();
            }
        });
        _undefinedTF.addFocusListener(new FocusAdapter() {
            @Override
            public void focusLost( FocusEvent e ) {
                regeneratePlot();

            }
        });
        _undefinedTF.setMaximumSize( _undefinedTF.getPreferredSize() );
        _undefinedTF.setToolTipText(resBundle.getString("undefinedToolTip"));
        Box undefBox = Box.createHorizontalBox();
        undefBox.add(undefinedLabel);
        undefBox.add(_undefinedTF);
        
		//	Create navigation buttons.
		JButton fullLeftBtn = new JButton(resBundle.getString("fullLeftBtnText"));
        fullLeftBtn.setToolTipText(resBundle.getString("fullLeftBtnToolTip"));
		fullLeftBtn.setHorizontalAlignment(SwingConstants.CENTER);
		fullLeftBtn.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doFullLeftButton();
			}
		} );
		
        _leftButton = new JButton(resBundle.getString("leftBtnText"));
        _leftButton.setToolTipText(resBundle.getString("leftBtnToolTip"));
		_leftButton.setEnabled(false);
		_leftButton.setHorizontalAlignment(SwingConstants.CENTER);
		_leftButton.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doLeftButton();
			}
		} );
		
        _rightButton = new JButton(resBundle.getString("rightBtnText"));
        _rightButton.setToolTipText(resBundle.getString("rightBtnToolTip"));
		_rightButton.setEnabled(_origTbl.getDimensions() > 2);
		_rightButton.setHorizontalAlignment(SwingConstants.CENTER);
		_rightButton.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doRightButton();
			}
		} );
			
		JButton fullRightBtn = new JButton(resBundle.getString("fullRightBtnText"));
        fullRightBtn.setToolTipText(resBundle.getString("fullRightBtnToolTip"));
		fullRightBtn.setHorizontalAlignment(SwingConstants.CENTER);
		fullRightBtn.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doFullRightButton();
			}
		} );
		
		JButton xPlotBtn = new JButton(resBundle.getString("xPlotBtnText"));
        xPlotBtn.setToolTipText(resBundle.getString("xPlotBtnToolTip"));
		xPlotBtn.setHorizontalAlignment(SwingConstants.CENTER);
		xPlotBtn.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				doXPlotButton();
			}
		} );
		
		JButton closeButton = new JButton(resBundle.getString("closeBtnText"));
		this.getRootPane().setDefaultButton(closeButton);
		closeButton.addActionListener(new ActionListener() {
            @Override
			public void actionPerformed(ActionEvent evt) {
				handleClose(null);
			}
		} );
		
		//	Create a button box and add all the buttons to it.
		Box btnBox = new Box(BoxLayout.X_AXIS);
		btnBox.setBackground( SystemColor.window );
        btnBox.add(undefBox);
		btnBox.add(Box.createHorizontalStrut(20));
		btnBox.add(Box.createGlue());

		btnBox.add(fullLeftBtn);
		btnBox.add(_leftButton);
		btnBox.add(_rightButton);
		btnBox.add(fullRightBtn);
		btnBox.add(Box.createHorizontalStrut(20));
		btnBox.add(xPlotBtn);
		btnBox.add(Box.createGlue());
		btnBox.add(Box.createHorizontalStrut(20));
		btnBox.add(closeButton);
		if (_origTbl.getDimensions() < 3) {
            fullLeftBtn.setEnabled(false);
            _leftButton.setEnabled(false);
            _rightButton.setEnabled(false);
            fullRightBtn.setEnabled(false);
		}
		if (_origTbl.getDimensions() < 2) {
            xPlotBtn.setEnabled(false);
		}
		
		JPanel buttonPanel = new JPanel(new BorderLayout());
		buttonPanel.add(btnBox, BorderLayout.SOUTH);
		buttonPanel.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
		
		_centerPanel.add(buttonPanel, BorderLayout.SOUTH);
		
		//	Create the plot panel containing the plot itself.
		ChartPanel plotPanel = createPlotPanel(origTable, interpTable, _nf, _index, this);
		_centerPanel.add(plotPanel, BorderLayout.CENTER);
		
		//	Create a higher order breakpoint panel if necessary.
		_bpPanel = createBPPanel(_origTbl, _nf, _index, _dimOrder);
		if (_bpPanel != null)
			_centerPanel.add(_bpPanel, BorderLayout.NORTH);
		
	}


	/**
	*  Method that will build an "Export" menu containing a list of all the FTableReader objects that
	*  are capable of writing to files.
	*
	*  @param  title	The title for the menu that is created.
	*  @returns A JMenu instance containing a list of all FTableReader objects that can write to
	*           files in various formats.  If there are no appropriate readers available,
	*           null is returned.
	*/
    @Override
	protected JMenu buildExportMenu(String title) {
		return null;
    }

    
	/**
	*  Method that regenerates and redraws the plot in the window.  The old
	*  plot is disposed of and a new plot and plot panel is created.
	*/
	private void regeneratePlot() {
		
		try {
			//	Remove the old plot.
			if (getChartPanel() != null)
				_centerPanel.remove(getChartPanel());
			
			//	Create the new plot and add it to the window.
			ChartPanel plotPanel = createPlotPanel(_origTbl, _interpTbl, _nf, _index, this);
			_centerPanel.add(plotPanel, BorderLayout.CENTER);
			
			//	Remove the high dimension breakpoint panel.
			if (_bpPanel != null)
				_centerPanel.remove(_bpPanel);

			//	Create a new breakpoint label at the top (if necissary).
			_bpPanel = createBPPanel(_origTbl, _nf, _index, _dimOrder);
			if (_bpPanel != null)
				_centerPanel.add(_bpPanel, BorderLayout.NORTH);
			
            //	Have the window re-configure itself.
			validate();
		
		} catch( NoSuchMethodException e ) {
			System.err.println( "Menu action listener method not found!" );
			e.printStackTrace();
		}

	}
	
	/**
	*  Creates a panel that contains the plot of the data.
	*
	* @param origTbl    The original un-interpolated table being plotted.
    * @param interpTbl  The interpolated table from ODIE.
	* @param nf         Number format used to format plot labels.
	* @param index      Pre-allocated storage space for indices into the plot table.
	* @param target     The object that receives the action events.  Must have the
	*                      following methods defined:  "handlePrint(ActionEvent e)", and
	*                      "handlePageSetup(ActionEvent e)".
	*/
	private ChartPanel createPlotPanel(Float64Table origTbl, Float64Table interpTbl, NumberFormat nf, List<Integer64> index, Object target) throws NoSuchMethodException {
	
		//	Create the chart.
        JFreeChart chart = createChart(origTbl, interpTbl, nf, index, _dimOrder, _undefinedTF.getValue());

		//	Create a panel to put the chart in.
		createChartPanel(chart, target);
		
		//	Return a reference to the newly created chart panel.
		return getChartPanel();
	}
	
	/**
	* Creates the plot of the table showing the 1st & 2nd independents at
	* the current breakpoint for higher independents.
	*
	* @param origTbl    The original un-interpolated table being plotted.
    * @param interpTbl  The interpolated table from ODIE.
	* @param  nf        Number format used to format plot labels.
	* @param  index     Pre-allocated storage space for indices into the plot table.
	* @param  dimOrder  List of integers that give the order that the dimensions are to be plotted.
    * @param undefined The value to use to represent an undefined number that shouldn't be plotted.
	*/
	private static JFreeChart createChart(Float64Table origTbl, Float64Table interpTbl, NumberFormat nf,
            List<Integer64> index, int[] dimOrder, double undefined) {
	
		// List of indexes into the n-dimensional table.
		int dim = origTbl.getDimensions();

		String indepName = origTbl.getIndepName(dimOrder[dim - 1]).toString();

		//	Build up a run list.
		XYSeriesCollection runList = new XYSeriesCollection();
		if (dim < 2) {
			//	Just a single run for 1D plots.
            Float64Vector xArr = origTbl.getBreakpoints(dimOrder[dim - 1]);
			runList.addSeries(createRun(origTbl, indepName, xArr, index, dimOrder, undefined));
            Float64Vector xArrInterp = interpTbl.getBreakpoints(dimOrder[dim-1]);
            runList.addSeries(createRun(interpTbl, indepName+"_odie", xArrInterp, index, dimOrder, Double.NaN));
			
		} else {
			//	All higher dimension plots are represented by a plot with multiple runs and panes.
			String indepName2 = origTbl.getIndepName(dimOrder[dim - 2]).toString();
			Float64Vector d2BrkPoints = origTbl.getBreakpoints(dimOrder[dim - 2]);
            
            //  Create the original (uninterpolated) set of curves.
            Float64Vector xArr = origTbl.getBreakpoints(dimOrder[dim - 1]);
			int numD2BrkPoints = d2BrkPoints.getDimension();
			for (int i=0; i < numD2BrkPoints; ++i) {
				index.set(dimOrder[dim - 2], Integer64.valueOf(i));			//	index[dimOrder[dim - 2]] = i;
				String runName = indepName2 + " = " + nf.format(d2BrkPoints.getValue(i));
				runList.addSeries(createRun(origTbl, runName, xArr, index, dimOrder, undefined));
			}
            
            //  Create the interpolated curves from the ODIE output table.
            for (int i=0; i < numD2BrkPoints; ++i) {
				index.set(dimOrder[dim - 2], Integer64.valueOf(i));			//	index[dimOrder[dim - 2]] = i;
                String runName = indepName2 + "_odie = " + nf.format(d2BrkPoints.getValue(i));
                runList.addSeries(createInterpRun(origTbl, interpTbl, runName, index, dimOrder, undefined));
            }
		}

		//	Create the chart.
        JFreeChart chart = ChartFactory.createXYLineChart(
												null,						//	No Title
												indepName,					//	Domain axis label
												origTbl.getName().toString(),	//	Range axis label
												runList,					//	Data to be plotted
												PlotOrientation.VERTICAL,	//	Horizontal orientation
												true,						//	Include legend
												true,						//	Include tooltips
												false);						//	Do not include URLs

		//	Move legend to the right hand side of the plot.
        LegendTitle legend = chart.getLegend();
		legend.setPosition(org.jfree.ui.RectangleEdge.RIGHT);
        
		//	Adjust the renderer features.
		XYPlot plot = chart.getXYPlot();
		XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(true,true);
		renderer.setBaseToolTipGenerator(new StandardXYToolTipGenerator());
		plot.setRenderer(renderer);
		
        //  Adjust the series rendering.
        BasicStroke origStroke = (BasicStroke)renderer.lookupSeriesStroke(0);
        float[] dashArr = {3F, 3F};
        Stroke dashedStroke = new BasicStroke(origStroke.getLineWidth(), origStroke.getEndCap(), origStroke.getLineJoin(), origStroke.getMiterLimit(), dashArr, 0F);
        if (dim < 2) {
            //  We have a 1D table.

            //  Make the original (uninterpolated) lines dashed.
            renderer.setSeriesStroke(0, dashedStroke);

            //  Turn off symbols for the interpolated table line & make its Color match the uninterpolated.
            renderer.setSeriesPaint(1, renderer.lookupSeriesPaint(0));
            renderer.setSeriesShapesVisible(1, false);

            //  Remove redundant data from the legend.
            renderer.setSeriesVisibleInLegend(1, false);

        } else {
            //  We have a 2D or greater table.
            
            Float64Vector d2BrkPoints = origTbl.getBreakpoints(dimOrder[dim - 2]);
            int numD2BrkPoints = d2BrkPoints.getDimension();
            
            for (int i=0; i < numD2BrkPoints; ++i) {
                //  Make the original (uninterpolated) lines dashed.
                renderer.setSeriesStroke(i, dashedStroke);
                
                //  Turn off symbols for the interpolated table line & make its Color match the uninterpolated.
                int i2 = i+numD2BrkPoints;
                renderer.setSeriesPaint(i2, renderer.lookupSeriesPaint(i));
                renderer.setSeriesShapesVisible(i2, false);
                
                //  Remove redundant data from the legend.
                renderer.setSeriesVisibleInLegend(i2, false);
            }
        }
		
		return chart;
	}


	/**
	*  Creates a single run from a table at the specified dimension.
	*/
	private static XYSeries createRun(Float64Table table, String runName, Float64Vector xArr, List<Integer64> index, int[] dimOrder, double undefined) {
	
		XYSeries run = new XYSeries(runName);
		int dim = table.getDimensions();
		int length = xArr.getDimension();
		for (int i=0; i < length; ++i) {
			index.set(dimOrder[dim - 1], Integer64.valueOf(i));			//	index[dimOrder[dim - 1]] = i;
			double depValue = table.getValue(index);
            if (Math.abs(depValue - undefined) < EPS)
                depValue = Double.NaN;
			run.add(xArr.getValue(i), depValue);
		}
		
		return run;
	}
	
	
	/**
	*  Creates a single run by interpolating the 2nd dimension breakpoint from a table at the specified dimension.
	*/
	private static XYSeries createInterpRun(Float64Table origTbl, Float64Table interpTbl, String runName, List<Integer64> index, int[] dimOrder, double undefined) {
	
        //  Pull out the original table breakpoints in each dimension (specified by "index").
		int numDims = interpTbl.getDimensions();
        double[] independents = new double[numDims];
        for (int dim=0; dim < numDims; ++dim) {
            double bpValue = origTbl.getBreakpointValue(dim, index.get(dim).intValue());
            independents[dim] = bpValue;
        }
        
        //  Interpolate into the interpTbl at the breakpoints from the original table except for the X-axis dimension
        //  where we are using the interpTbl breakpoints.
        Float64Vector xArr = interpTbl.getBreakpoints(dimOrder[numDims-1]);
		XYSeries run = new XYSeries(runName);
		int length = xArr.getDimension();
		for (int i=0; i < length; ++i) {
            double xBPValue = xArr.getValue(i);
            independents[dimOrder[numDims - 1]] = xBPValue;     //  Put in the X-axis breakpoint value in the correct table dimension.
			double depValue = interpTbl.lookupValue(independents);
            if (Math.abs(depValue - undefined) < EPS)
                depValue = Double.NaN;
			run.add(xArr.getValue(i), depValue);
		}
		
		return run;
	}
	
	
	
	/**
	*  Create a high dimension breakpoint panel.  These are for breakpoints shown as different plot panes.
	*/
	private static JPanel createBPPanel(Float64Table table, NumberFormat nf, List<Integer64> index, int[] dimOrder) {
		JPanel thePanel = null;
		
		int dim = table.getDimensions();
		if (dim > 2) {
			int numElements = dim - 2;
			thePanel = new JPanel(new GridLayout(numElements, 1));
			
			for (int i=0; i < numElements; ++i) {
				Float64Vector breakpoints = table.getBreakpoints(dimOrder[i]);
				String indepName = table.getIndepName(dimOrder[i]).toString();
				double bpValue = breakpoints.getValue(index.get(dimOrder[i]).intValue());
				String message = indepName + " = " + nf.format(bpValue);
				JLabel label = new JLabel(message, SwingConstants.CENTER);
				thePanel.add(label);
			}
		}
		
		return thePanel;
	}
	
	/**
	*  Handles the user clicking on the full-left button.
	*/
	private void doFullLeftButton() {
		_leftButton.setEnabled(false);
		_rightButton.setEnabled(true);
		
		int dim = _origTbl.getDimensions() - 3;
		for (int i=0; i <= dim; ++i) {
			_index.set(_dimOrder[i], Integer64.ZERO);			//	_index[dimOrder[i]] = 0;
		}

		regeneratePlot();
	}

	/**
	*  Handles the user clicking on the left button.
	*/
	private void doLeftButton() {
		_rightButton.setEnabled(true);
		int dim = _origTbl.getDimensions() - 3;
		int old_dim = dim;
		
		_index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).minus(Integer64.ONE));		//	--_index[dimOrder[dim]];
		while (dim != 0 && _index.get(_dimOrder[dim]).intValue() < 0) {
			int bpIdx = _origTbl.getNumBreakpoints(_dimOrder[dim]) - 1;
			_index.set(_dimOrder[dim], Integer64.valueOf(bpIdx));						//	_index[dimOrder[dim]] = bpIdx;
			--dim;
			_index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).minus(Integer64.ONE));	//	--_index[dimOrder[dim]];
		}
		
		//	Have we reached the end of all the breakpoints of
		//	all the dimensions of the table?
		dim = old_dim;
		while (dim != 0 && _index.get(_dimOrder[dim]).intValue() == 0) {
			--dim;
		}
		if (dim == 0 && _index.get(_dimOrder[dim]).intValue() == 0)
			_leftButton.setEnabled(false);
		
		regeneratePlot();
	}
	
	/**
	*  Handles the user clicking on the right button.
	*/
	private void doRightButton() {
		_leftButton.setEnabled(true);
		int dim = _origTbl.getDimensions() - 3;
		int old_dim = dim;
		
		_index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).plus(Integer64.ONE));		//	++_index[dimOrder[dim]];
		while ( dim != 0 && _index.get(_dimOrder[dim]).intValue() == _origTbl.getNumBreakpoints(_dimOrder[dim]) ) {
			_index.set(_dimOrder[dim], Integer64.ZERO);			//	_index[dimOrder[dim]] = 0;
			--dim;
			_index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).plus(Integer64.ONE));	//	++_index[dimOrder[dim]];
		}
		
		//	Have we reached the end of all the breakpoints of
		//	all the dimensions of the table?
		dim = old_dim;
		while (dim != 0 && _index.get(_dimOrder[dim]).intValue() == _origTbl.getNumBreakpoints(_dimOrder[dim]) - 1) {
			--dim;
		}
		if (dim == 0 && _index.get(_dimOrder[dim]).intValue() == _origTbl.getNumBreakpoints(_dimOrder[dim]) - 1)
			_rightButton.setEnabled(false);
		
		regeneratePlot();
	}
	
	/**
	*  Handles the user clicking on the full-right button.
	*/
	private void doFullRightButton() {
		_leftButton.setEnabled(true);
		_rightButton.setEnabled(false);
		
		int dim = _origTbl.getDimensions() - 3;
		for (int i=0; i <= dim; ++i) {
			int bpIdx = _origTbl.getNumBreakpoints(_dimOrder[i]) - 1;
			_index.set(_dimOrder[i], Integer64.valueOf(bpIdx));			//	_index[dimOrder[i]] = bpIdx;
		}

		regeneratePlot();
	}

	/**
	*  Handles the user clicking on the X-Plot button.
	*/
	private void doXPlotButton() {
		
		if (_origTbl.getDimensions() == 2) {
			int tmp = _dimOrder[0];
			_dimOrder[0] = _dimOrder[1];
			_dimOrder[1] = tmp;
			
		} else {
		
			//  If there is more than two dimensions, 
			//	ask the user to select the order of the dimensions to plot.
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
			SelectPlotDimsDialog dimDialog = new SelectPlotDimsDialog(this, resBundle.getString("selectPlotDimsTitle"),
													resBundle.getString("selectPlotDimsText"), _origTbl, _dimOrder);
			dimDialog.setLocation(AppUtilities.dialogPosition(dimDialog));
			dimDialog.setVisible(true);

			//	Retrieve the selected cases.
			List dimensions = dimDialog.getSelected();
			dimDialog.dispose();
			if (dimensions == null)	return;		//	User canceled.
			
			//	Copy the output from the dialog to the dimensions order array.
			int dims = _dimOrder.length;
			for (int i=0; i < dims; ++i) {
				Integer element = (Integer)dimensions.get(i);
				_dimOrder[i] = element;
			}
			
			//	Reset the plot to the left end of the 3rd independent variable.
			doFullLeftButton();
			
		}

		//	Draw the plot with the new order of dimensions.
		regeneratePlot();
	}

}

