/*
*  TableViewWindow -- Window that displays an n-dimensional table in tabular form.
*
*  Copyright (C) 2000-2024, Joseph A. Huwaldt. All rights reserved.
*
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
package jahuwaldt.jodie;

import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.js.tables.OTISTableReader;
import jahuwaldt.swing.*;
import jahuwaldt.util.GeneralFormat;
import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.*;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.List;
import java.util.ResourceBundle;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Integer64;

/**
 * A window for displaying the contents of an n-dimensional functional table in tabular
 * (spreadsheet) form.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: June 27, 2002
 * @version January 19, 2024
 */
public class TableViewWindow extends JFrame {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(TableViewWindow.class.getName());

    //  Table meta-data keys.
    private static final String CURVEFITDEGREE = OTISTableReader.CURVEFITDEGREE_KEY;
    private static final String CHAMFERWIDTH = OTISTableReader.CHAMFERWIDTH_KEY;

    //  An adapter that will place our table data on the clipboard in an
    //  Excel compatible format.
    private final TableClipboardAdapter clipboard;

    //	Number formats used to parse/format text strings.
    private NumberFormat nfd = new GeneralFormat();
    private NumberFormat nfi = new GeneralFormat();

    //	The table to be displayed in this window.
    private Float64Table table;

    //	Names of the independent and dependent variables in the table.
    private String[] names;

    //	Array used to set position in a multi-dimensional table.
    private List<Integer64> _pos;

    //  Text fields & labels for the chamfer width.
    private DecimalField[] chamferFields;
    private JLabel[] chamferLabels;

    /**
     * Constructs a new table view window that displays the specified table in tabular
     * form.
     *
     * @param application Reference to the main application class.
     * @param theTable    The table we want to do a table lookup on.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public TableViewWindow(Float64Table tbl) throws NoSuchMethodException {
        super(tbl.toString());

        //  Have the window dispose of itself when it closes.
        this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

        this.setResizable(true);

        //  Define the window icon in the taskbar.
        ImageIcon icon = AppUtilities.getResourceIcon(MDIApplication.getInstance().getResourceBundle().getString("applicationIconResName"), LOGGER);
        if (icon != null)
            this.setIconImage(icon.getImage());

        //	Save off the table.
        this.table = tbl;

        //  Create a panel to contain interpolation information.
        JPanel interpPanel = new JPanel(new SpringLayout());
        interpPanel.setBorder(BorderFactory.createTitledBorder("Interpolation:"));
        String[] cfOptions = {"Floor", "Linear", "Cubic", "Quintic", "Chamfer"};

        //  Create a label, combobox, label and text field for each independent variable.
        int numDims = tbl.getDimensions();
        chamferFields = new DecimalField[numDims];
        chamferLabels = new JLabel[numDims];
        for (int dim = 0; dim < numDims; ++dim) {
            //  Get the name of each independent variable and add it as a label.
            String indepName = tbl.getIndepName(dim).toString();
            JLabel label = new JLabel(indepName, JLabel.TRAILING);
            interpPanel.add(label);

            //  Get the OTIS interpolation type.
            indepName = OTISTableReader.sanitizeIndepName(indepName).toString();
            String key = CURVEFITDEGREE + "_" + indepName;
            Integer cfDeg = (Integer) tbl.getMetaData(key);
            if (cfDeg == null)
                cfDeg = 5;

            //  Create a combo-box for the interpolation type.
            MyComboBoxModel cbModel = new MyComboBoxModel(dim, cfOptions);
            JComboBox cb = new JComboBox(cbModel);
            switch (cfDeg) {
                case 0:
                case 1:
                    cb.setSelectedIndex(cfDeg);
                    break;
                case 3:
                    cb.setSelectedIndex(2);
                    break;
                case 5:
                    cb.setSelectedIndex(3);
                    break;
                case 6:
                    cb.setSelectedIndex(4);
                    break;
            }
            interpPanel.add(cb);

            //  Set up an action item listener for the combo box.
            cb.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    JComboBox cb = (JComboBox) e.getSource();
                    int dim = ((MyComboBoxModel) cb.getModel()).getDimension();
                    int sel = cb.getSelectedIndex();

                    //  Determine the OTIS curve fit degree from the user selection.
                    int cfDeg = 5;
                    switch (sel) {
                        case 0:
                        case 1:
                            cfDeg = sel;
                            break;
                        case 2:
                            cfDeg = 3;
                            break;
                        case 3:
                            cfDeg = 5;
                            break;
                        case 4:
                            cfDeg = 6;
                            break;
                    }

                    //  Get the name of the independent variable in this dimension.
                    String indepName = OTISTableReader.sanitizeIndepName(table.getIndepName(dim)).toString();

                    //  Store the new curve fit degree in the table's meta data.
                    String key = CURVEFITDEGREE + "_" + indepName;
                    table.putMetaData(key, cfDeg);

                    key = CHAMFERWIDTH + "_" + indepName;
                    table.removeMetaData(key);
                    if (cfDeg == 6) {
                        //  Handle chamfer.
                        //  Enable the associated chamfer width.
                        chamferLabels[dim].setEnabled(true);
                        DecimalField tf = chamferFields[dim];
                        tf.setEnabled(true);
                        double Wc = tf.getValue();
                        table.putMetaData(key, Wc);
                    }
                }
            });

            //  Create a label and text field for the chamfer width.
            label = new JLabel("width = ", JLabel.TRAILING);
            DecimalField tf;
            if (cfDeg == 6) {
                //  Chamfer needs a chamfer width.
                key = CHAMFERWIDTH + "_" + indepName;
                Double Wc = (Double) tbl.getMetaData(key);
                if (Wc == null)
                    Wc = 0.05;

                tf = new DecimalField(Wc, 8, new GeneralFormat());

            } else {
                //  Just put in a blank entry.
                label.setEnabled(false);
                tf = new DecimalField(0.05, 8, new GeneralFormat());
                tf.setEnabled(false);
            }
            interpPanel.add(label);
            interpPanel.add(tf);
            chamferFields[dim] = tf;
            chamferLabels[dim] = label;

            //  Add an action listener and focus listener for the chamfer width text field.
            tf.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    DecimalField tf = (DecimalField) e.getSource();
                    int dim = 0;
                    for (; dim < table.getDimensions(); ++dim) {
                        if (chamferFields[dim] == tf)
                            break;
                    }
                    String indepName = OTISTableReader.sanitizeIndepName(table.getIndepName(dim)).toString();
                    String key = CHAMFERWIDTH + "_" + indepName;
                    double Wc = tf.getValue();
                    table.putMetaData(key, Wc);
                }
            });
            tf.addFocusListener(new FocusAdapter() {
                @Override
                public void focusLost(FocusEvent e) {
                    DecimalField tf = (DecimalField) e.getSource();
                    int dim = 0;
                    for (; dim < table.getDimensions(); ++dim) {
                        if (chamferFields[dim] == tf)
                            break;
                    }
                    String indepName = OTISTableReader.sanitizeIndepName(table.getIndepName(dim)).toString();
                    String key = CHAMFERWIDTH + "_" + indepName;
                    double Wc = tf.getValue();
                    table.putMetaData(key, Wc);
                }
            });

        }
        //  Layout the interpolation panel.
        SpringUtilities.makeCompactGrid(interpPanel,
                numDims, 4, //rows, cols
                0, 0, //initX, initY
                6, 6);       //xPad, yPad

        //  Allocate memory for some arrays.
        names = new String[numDims + 1];

        _pos = FastTable.newInstance();
        for (int i = 0; i < numDims; ++i)
            _pos.add(Integer64.ZERO);

        //	Create a column for each independent variable followed by the dependent.
        for (int i = 0; i < numDims; ++i)
            names[i] = tbl.getIndepName(i).toString();
        names[numDims] = tbl.getName().toString();

        // Create a table to show the data.
        final DecimalField decField = new DecimalField(8, nfd);
        DefaultCellEditor decEditor = new DefaultCellEditor(decField) {
            //Override DefaultCellEditor's getCellEditorValue method
            //to return an Float, not a String:
            @Override
            public Object getCellEditorValue() {
                return nfd.format(decField.getValue());
            }
        };
        TableModel dataModel = new ViewTableModel();
        JTable tableView = new JTable(dataModel);
        tableView.setDefaultEditor(String.class, decEditor);
        tableView.setCellSelectionEnabled(true);
        clipboard = new TableClipboardAdapter(tableView);
        JScrollPane scrollpane = new JScrollPane(tableView);
        scrollpane.setPreferredSize(new Dimension(250, 300));
        scrollpane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

        //	Create a border for the scroll pane.
        Border border = scrollpane.getBorder();
        border = BorderFactory.createCompoundBorder(new EmptyBorder(0, 0, 12, 0), border);
        scrollpane.setBorder(border);

        //	Layout the main window.
        Container cp = getContentPane();
        cp.add(interpPanel, BorderLayout.NORTH);
        cp.add(scrollpane, BorderLayout.CENTER);

        //  Create a menu bar for this window.
        JMenuBar menuBar = createMenuBar();
        this.setJMenuBar(menuBar);

        //  Size the window.
        this.pack();

        //  Position the window so that each new window can be seen.
        AppUtilities.positionWindow(this, getWidth(), getHeight());
    }

    /**
     * A combo-box model that tracks the table dimension associated with the combo-box.
     */
    private class MyComboBoxModel extends DefaultComboBoxModel {

        private final int _dim;

        public MyComboBoxModel(int dim, Object[] items) {
            super(items);
            _dim = dim;
        }

        public int getDimension() {
            return _dim;
        }
    }

    /**
     * Initializes and displays the menus associated with this application.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        MDIApplication guiApp = MDIApplication.getInstance();
        ResourceBundle resBundle = guiApp.getResourceBundle();

        //  Create a menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        int row = 0;
        String[][] menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("newItemText");
        menuStrings[row][1] = resBundle.getString("newItemKey");
        menuStrings[row][2] = "handleNew";
        ++row;
        menuStrings[row][0] = resBundle.getString("openItemText");
        menuStrings[row][1] = resBundle.getString("openItemKey");
        menuStrings[row][2] = null;
        ++row;
        ++row;	//	Blank line
        menuStrings[row][0] = resBundle.getString("closeItemText");
        menuStrings[row][1] = resBundle.getString("closeItemKey");
        menuStrings[row][2] = "handleClose";
        ++row;
        menuStrings[row][0] = resBundle.getString("saveItemText");
        menuStrings[row][1] = resBundle.getString("saveItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("saveAsItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;

        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }

        
        // Set up the edit menu.
        row = 0;
        menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("undoItemText");
        menuStrings[row][1] = resBundle.getString("undoItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("redoItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        ++row;	//	Blank line.
        menuStrings[row][0] = resBundle.getString("cutItemText");
        menuStrings[row][1] = resBundle.getString("cutItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("copyItemText");
        menuStrings[row][1] = resBundle.getString("copyItemKey");
        menuStrings[row][2] = "handleCopy";
        ++row;
        menuStrings[row][0] = resBundle.getString("pasteItemText");
        menuStrings[row][1] = resBundle.getString("pasteItemKey");
        menuStrings[row][2] = "handlePaste";

        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

        //  Create a Preferences menu item (if not automatically present).
        if (!PreferencesJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getPreferencesJMenuItem());
        }

		//  Create a Window's menu.
		menu = MDIApplication.newWindowsMenu(resBundle.getString("windowsMenuText"));
		menuBar.add(menu);
		
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);

            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }

        return menuBar;
    }

    /**
     * Handles the user choosing "New..." from the File menu. Creates a new application
     * window instance.
     * 
     * @param event The action that caused this method to be called.
     */
    public void handleNew(ActionEvent event) {
        MDIApplication.getInstance().handleNew(event);
    }

    /**
     * Handle the user choosing "Close" from the File menu. This implementation dispatches
     * a "Window Closing" event to this window.
     * 
     * @param event The action that caused this method to be called.
     */
    public void handleClose(ActionEvent event) {
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

    /**
     * Handles the user choosing "Copy" from the Edit menu.
     * 
     * @param event The action that caused this method to be called.
     */
    public void handleCopy(ActionEvent event) {
        clipboard.copy();
    }

    /**
     * Handles the user choosing "Paste" from the Edit menu.
     * 
     * @param event The action that caused this method to be called.
     */
    public void handlePaste(ActionEvent event) {
        try {
            clipboard.paste();
        } catch (Exception e) {
            //  Simply ignore any attempt to paste bad data.
        }
    }

    /**
     * A model for our table that figures out what value to put in what cell.
     */
    private class ViewTableModel extends AbstractTableModel {

        // These methods always need to be implemented.
        @Override
        public int getColumnCount() {
            return names.length;
        }

        @Override
        public int getRowCount() {
            return table.size();
        }

        @Override
        public Object getValueAt(int row, int col) {
            String value;

            int numDims = table.getDimensions();
            if (col == numDims) {
                //  We are dealing with the dependent value.

                //  Determine position in the table.
                _pos = tablePosition(row, numDims, _pos);

                //  Get the value from the table.
                value = nfd.format(table.getValue(_pos));

            } else {
                //  We are dealing with one of the independents.

                //  Determine position in the table.
                int idx = indepPosition(row, col);

                //  Get the value from the table.
                value = nfi.format(table.getBreakpoint(col, idx));
            }

            return value;
        }

        // The default implementations of these methods in
        // AbstractTableModel would work, but we can refine them.
        @Override
        public String getColumnName(int column) {
            return names[column];
        }

        @Override
        public Class getColumnClass(int col) {
            return getValueAt(0, col).getClass();
        }

        @Override
        public boolean isCellEditable(int row, int col) {
            boolean value = true;
            if (col < table.getDimensions())
                value = false;
            return value;
        }

        @Override
        public void setValueAt(Object aValue, int row, int col) {
            try {

                int numDims = table.getDimensions();
                if (col == numDims) {
                    //  We are dealing with the dependent value.

                    double value = nfd.parse((String) aValue).doubleValue();

                    //  Determine position in the table.
                    _pos = tablePosition(row, numDims, _pos);

                    //  Set the dependent value.
                    table.set(_pos, value);

                } else {
                    //  We are dealing with one of the independents.

                    double value = nfi.parse((String) aValue).doubleValue();

                    //  Determine position in the table.
                    int idx = indepPosition(row, col);

                    //  Set the independent value.
                    table.setBreakpoint(col, idx, value);
                }

            } catch (ParseException e) {
                ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                JOptionPane.showMessageDialog(TableViewWindow.this, resBundle.getString("illegalInputMsg"),
                        resBundle.getString("parseErrorTitle"), JOptionPane.ERROR_MESSAGE);
                TableViewWindow.this.repaint();
            }

        }

        private List<Integer64> tablePosition(int row, int numDims, List<Integer64> pos) {

            int numElements = table.size();
            for (int i = 0; i < numDims; ++i) {
                numElements /= table.getNumBreakpoints(i);

                int idx = 0;
                if (numElements > 1) {
                    int diff = row - numElements;
                    while (diff >= 0) {
                        diff -= numElements;
                        ++idx;
                    }

                } else
                    idx = row;
                _pos.set(i, Integer64.valueOf(idx));			//	_pos[i] = idx;
                row -= idx * numElements;
            }

            return pos;
        }

        /**
         * Determine the index into the specified independent variable dimension that
         * corresponds to the specified row in the table.
         */
        private int indepPosition(int row, int dimension) {
            int numDim = table.getDimensions();

            int pos = recursivepos(row, dimension, numDim, 0, 0);

            return pos;
        }

        private int recursivepos(int row, int tdim, int numDim, int dim, int pposf) {

            int denom = 1;
            for (int i = dim + 1; i < numDim; ++i)
                denom *= table.getNumBreakpoints(i);

            int pos = (row - pposf) / denom;
            if (dim != tdim)
                pos = recursivepos(row, tdim, numDim, ++dim, pposf + pos * denom);

            return pos;
        }

    }

}
