/*
 *   POSTTableReader -- A class that can read and write POST table (*.dat) multi-table files.
 *
 *   Copyright (C) 2008-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import static jahuwaldt.js.tables.FTableReader.RESOURCES;
import static java.io.StreamTokenizer.TT_EOF;
import static java.io.StreamTokenizer.TT_NUMBER;
import static java.io.StreamTokenizer.TT_WORD;
import java.io.*;
import java.text.DateFormat;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.NoSuchElementException;
import static java.util.Objects.requireNonNull;
import java.util.logging.Level;
import java.util.logging.Logger;
import javolution.text.Text;
import javolution.text.TextBuilder;
import javolution.text.TypeFormat;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A class that provides methods for reading or writing a POST table (*.dat) formatted
 * multi-table file. These files are used by the POST trajectory simulation system to
 * store aerodynamic and propulsion data among other things. See POST documentation for
 * more information.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: December 8, 2008
 * @version January 15, 2024
 */
public class POSTTableReader implements FTableReader {

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "dat";

    private static final Text SPACE = Text.intern(" ");

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("postDescription");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return The file extension "dat".
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.YES if the stream is recognized as an POST table
     * file.
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.YES if
     *         the file appears to contain POST tables.
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //	Mark the current position in the input stream so that we can return to it.
        input.mark(10240);	//	10KB buffer

        try {
            //	Create a buffer to hold the data read in from the file.
            byte[] buffer = new byte[10238];

            //	Read in up to 10k worth of data.
            int byteCount = input.read(buffer, 0, 10238);
            if (byteCount < 0)
                return NO;

            //	Turn the buffer into a (potentially 10k long) string.
            String str = new String(buffer, 0, byteCount);
            str = preProcessesData(str);

            //	Convert the String into a stream tokenizer.
            StreamTokenizer tokenizer = createTokenizer(str);

            //	Look for the "$tab" sequence.
            int ttype = findTable(tokenizer);
            if (ttype == TT_WORD)
                return YES;

        } finally {
            input.reset();
        }

        return NO;
    }

    /**
     * Returns true. This class can write data to an POST table file.
     */
    @Override
    public boolean canWriteData() {
        return true;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        requireNonNull(o, MessageFormat.format(RESOURCES.getString("paramNullErr"), "o"));
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in POST table formatted data from the specified input stream and
     * returns that data as an {@link FTableList} object.
     *
     * @param input An input stream containing the POST table formatted data.
     * @return An {@link FTableList} object that contains the data read in from the
     *         specified stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));

        //  POST II files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            //	Read the input stream into a String buffer.
            String data = readAll2String(input);

            //	Create a StreamTokenizer that reads from the String.
            StreamTokenizer tokenizer = createTokenizer(data);

            //	Parse the data.
            FTableList db = readStream(tokenizer);

            return db;

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Read in the entire file and return a a String which represents a buffer of the
     * data.
     */
    private String readAll2String(InputStream input) throws IOException {

        int size = input.available();

        //	Create a buffer to hold the data read in from the file.
        byte[] buffer = new byte[size * 2];

        //	Read in the data to a byte buffer.
        int byteCount = input.read(buffer, 0, size * 2);
        if (byteCount < 0)
            return "";

        //	Convert the byte buffer to a String.
        String str = new String(buffer, 0, byteCount);
        str = preProcessesData(str);

        return str;
    }

    /**
     * Method for writing out all the data stored in the specified list of
     * {@link Float64Table} objects to the specified output stream in POST table format.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of <code>Float64Table</code> objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        requireNonNull(output, MessageFormat.format(RESOURCES.getString("paramNullErr"), "output"));
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        if (tables.isEmpty())
            return;

        //  POST II files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Get a reference to the output stream.
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(output));

            //	Write out the comments.
            writeFileComments(writer, tables);

            //	Write out the list of tables.
            for (Float64Table table : tables) {
                List<Text> notes = table.getAllNotes();
                for (Text note : notes) {
                    writer.write("c ");
                    note.println(writer);
                }
                writer.write(" $");
                writer.newLine();
                writer.write("p$tab");
                writer.newLine();
                writeTable(writer, table);
            }

            //	Flush the output.
            writer.flush();

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Read in a POST table formatted multi-table stream and create a new table database
     * with the data.
     *
     * @param tokenizer The tokenizer for the stream.
     * @return A table database containing the tables read in from the file.
     */
    private static FTableList readStream(StreamTokenizer tokenizer) throws IOException {

        // Create an empty table database.
        FTableList db = FTableList.newInstance();

        try {
            int ttype = tokenizer.nextToken();
            while (ttype != TT_EOF) {

                //	Look for the "$tab table" sequence.
                while (ttype != TT_EOF) {
                    String sval = tokenizer.sval;
                    if (ttype == TT_WORD && ("$tab".equals(sval) || "p$tab".equals(sval))) {
                        ttype = tokenizer.nextToken();
                        sval = tokenizer.sval;
                        if (ttype == TT_WORD && "table".equals(sval))
                            break;
                    }
                    ttype = tokenizer.nextToken();
                }
                if (ttype == TT_EOF)
                    break;
                if (ttype != TT_WORD)
                    throw new IOException(RESOURCES.getString("postMissingTabSeqErr"));

                //	Skip the equals sign.
                ttype = tokenizer.nextToken();
                if (ttype != '=')
                    throw new IOException(format(RESOURCES.getString("p2MissingEqualSignErr"), tokenizer.lineno()));

                //	Get the table name.
                ttype = tokenizer.nextToken();
                if (ttype != TT_WORD && ttype != '\'' && ttype != '"')
                    throw new IOException(format(RESOURCES.getString("p2CantParseTblNameErr"), tokenizer.lineno()));
                String tblName = tokenizer.sval.toUpperCase();

                //	Get the number of dimensions.
                ttype = tokenizer.nextToken();
                if (ttype != TT_NUMBER)
                    throw new IOException(format(RESOURCES.getString("postMissingTblDimErr"), tokenizer.lineno()));
                int numDim = (int)tokenizer.nval;
                if (numDim < 0 || numDim > 3)
                    throw new IOException(format(RESOURCES.getString("postWrongNumBPErr"), tokenizer.lineno()));

                if (numDim > 0) {
                    //	Get the independent variable names.
                    FastTable<String> indepNames = FastTable.newInstance();
                    for (int i = 0; i < numDim; ++i) {
                        ttype = tokenizer.nextToken();
                        if (ttype != TT_WORD && ttype != '\'' && ttype != '"')
                            throw new IOException(format(RESOURCES.getString("postMissingIndepVarNameErr"), tokenizer.lineno()));
                        String varName = tokenizer.sval.toUpperCase();
                        if (varName == null || varName.length() == 0)
                            throw new IOException(format(RESOURCES.getString("postCantParseIndepVarNameErr"), tokenizer.lineno()));
                        indepNames.add(varName);
                    }

                    //	Get the number of breakpoints in each dimension.
                    int[] numBP = new int[numDim];
                    for (int i = 0; i < numDim; ++i) {
                        ttype = tokenizer.nextToken();
                        if (ttype != TT_NUMBER)
                            throw new IOException(format(RESOURCES.getString("postMissingNumBP4DimErr"), i + 1, tokenizer.lineno()));
                        int num = (int)tokenizer.nval;
                        if (num < 1)
                            throw new IOException(format(RESOURCES.getString("postWrongNumBP4DimErr"), i + 1, tokenizer.lineno()));
                        numBP[i] = num;
                    }

                    //	Skip the interpolation type.
                    ttype = tokenizer.nextToken();
                    if (ttype != TT_NUMBER)
                        throw new IOException(format(RESOURCES.getString("postCantParseInterpTypeErr"), tokenizer.lineno()));

                    //	Read in the direction of the values.
                    for (int i = 0; i < numDim; ++i) {
                        ttype = tokenizer.nextToken();
                        if (ttype != TT_NUMBER)
                            throw new IOException(format(RESOURCES.getString("postCantParseIndepDirErr"), indepNames.get(i), tokenizer.lineno()));
                        int direction = (int)tokenizer.nval;
                        if (direction != 1)
                            throw new IOException(format(RESOURCES.getString("postTblDecrErr"), tblName, indepNames.get(i)));
                    }

                    //	Skip the initial value of the pointers.
                    int skip = 1;
                    if (numDim == 2)
                        skip = 5;
                    if (numDim == 3)
                        skip = 21;
                    for (int i = 0; i < skip; ++i)
                        ttype = tokenizer.nextToken();

                    //	Read in the table based on the number of dimensions.
                    switch (numDim) {
                        case 1:
                            db.add(read1DTable(tokenizer, tblName, indepNames.get(0), numBP[0]));
                            break;
                        case 2:
                            db.add(read2DTable(tokenizer, tblName, indepNames, numBP));
                            break;
                        case 3:
                            db.add(read3DTable(tokenizer, tblName, indepNames, numBP));
                            break;
                    }

                    FastTable.recycle(indepNames);

                } else
                    Logger.getLogger(POSTTableReader.class.getName()).log(Level.WARNING,
                            format(RESOURCES.getString("postZeroDimMsg"), tblName));

            }	//	end while

        } catch (NoSuchElementException e) {
            throw new IOException(format(RESOURCES.getString("parseErrMsg"), tokenizer.lineno()), e);
        }

        return db;
    }

    /**
     * Method that reads in and returns a 1D table from the input stream.
     *
     * @param tokenizer A tokenizer that returns tokens from the input stream.
     * @return A table containing the 1D table data read in from the file.
     */
    private static Float64Table read1DTable(StreamTokenizer tokenizer, String tblName, String indepName, int numBP)
            throws IOException, NumberFormatException, NoSuchElementException {

        //	Create an independent and dependent variable array.
        FastTable<Float64> indepArr = FastTable.newInstance();
        FastTable<Float64> depArray = FastTable.newInstance();

        //	Begin reading in data by looping over the breakpoints.
        for (int i = 0; i < numBP; ++i) {

            //	Get the independent variable value.
            double value = TRUtils.nextNumber(tokenizer);
            indepArr.add(Float64.valueOf(value));

            //	Get the dependent value.
            value = TRUtils.nextNumber(tokenizer);
            depArray.add(Float64.valueOf(value));
        }

        Float64Table table = Float64Table.valueOf(tblName, indepName,
                Float64Vector.valueOf(indepArr), Float64Vector.valueOf(depArray));

        //	Clean up before leaving.
        FastTable.recycle(indepArr);
        FastTable.recycle(depArray);

        return table;
    }

    /**
     * Method that reads in and returns a 2D table from the input stream.
     *
     * @param tokenizer A tokenizer that returns tokens from the input stream.
     * @return A table containing the 2D table data read in from the file.
     */
    private static Float64Table read2DTable(StreamTokenizer tokenizer, String tblName, List<String> indepNames, int[] numBP)
            throws IOException, NumberFormatException, NoSuchElementException {

        //	Reverse the independent variables.
        int tmp = numBP[0];
        numBP[0] = numBP[1];
        numBP[1] = tmp;
        String tmpStr = indepNames.get(0);
        indepNames.set(0, indepNames.get(1));
        indepNames.set(1, tmpStr);

        //	Create a dependent variable array.
        FastTable<Float64Vector> depArray = FastTable.newInstance();

        //	Create the independent variable arrays.
        FastTable<Float64> indepArr1 = FastTable.newInstance();
        FastTable<Float64> indepArr2 = FastTable.newInstance();

        //	Begin reading in data by looping over the breakpoints.
        for (int i = 0; i < numBP[0]; ++i) {

            //	Get the independent value.
            double value = TRUtils.nextNumber(tokenizer);
            indepArr1.add(Float64.valueOf(value));

            FastTable<Float64> row = FastTable.newInstance();
            for (int j = 0; j < numBP[1]; ++j) {

                //	Get the independent variable value.
                value = TRUtils.nextNumber(tokenizer);
                Float64 f64Value = Float64.valueOf(value);
                if (i == 0)
                    indepArr2.add(f64Value);
                else if (!indepArr2.get(j).equals(f64Value))
                    throw new IOException(format(RESOURCES.getString("postVarBreakpointsErr"), tblName, tokenizer.lineno()));

                //	Get the dependent value.
                value = TRUtils.nextNumber(tokenizer);
                row.add(Float64.valueOf(value));
            }
            depArray.add(Float64Vector.valueOf(row));
            FastTable.recycle(row);
        }

        //	Create the table.
        Float64Table table = Float64Table.valueOf(tblName, indepNames,
                Float64Vector.valueOf(indepArr1), Float64Vector.valueOf(indepArr2), depArray);

        //	Clean up before leaving.
        FastTable.recycle(indepArr1);
        FastTable.recycle(indepArr2);
        FastTable.recycle(depArray);

        return table;
    }

    /**
     * Method that reads in and returns a 3D table from the input stream.
     *
     * @param tokenizer A tokenizer that returns tokens from the input stream.
     * @return A table containing the 3D table data read in from the file.
     */
    private static Float64Table read3DTable(StreamTokenizer tokenizer, String tblName, List<String> indepNames, int[] numBP)
            throws IOException, NumberFormatException, NoSuchElementException {

        //	Reverse the independent variables.
        int tmp = numBP[0];
        numBP[0] = numBP[2];
        numBP[2] = tmp;
        String tmpStr = indepNames.get(0);
        indepNames.set(0, indepNames.get(2));
        indepNames.set(2, tmpStr);

        //	Create a dependent variable array.
        FastTable<List<Float64Vector>> depArray = FastTable.newInstance();
        for (int i = 0; i < numBP[0]; ++i) {
            FastTable<Float64Vector> col = FastTable.newInstance();
            depArray.add(col);
        }

        //	Create the independent variable arrays.
        FastTable<Float64> indepArr1 = FastTable.newInstance();
        FastTable<Float64> indepArr2 = FastTable.newInstance();
        FastTable<Float64> indepArr3 = FastTable.newInstance();

        //	Begin reading in data by looping over the breakpoints.
        for (int i = 0; i < numBP[0]; ++i) {

            //	Get the independent value.
            double value = TRUtils.nextNumber(tokenizer);
            indepArr1.add(Float64.valueOf(value));

            for (int j = 0; j < numBP[1]; ++j) {

                //	Get the independent variable value.
                value = TRUtils.nextNumber(tokenizer);
                Float64 f64Value = Float64.valueOf(value);
                if (i == 0)
                    indepArr2.add(f64Value);
                else if (!indepArr2.get(j).equals(f64Value))
                    throw new IOException(format(RESOURCES.getString("postIncIndepValuesErr"), tblName, tokenizer.lineno()));

                FastTable<Float64> rank = FastTable.newInstance();
                for (int k = 0; k < numBP[2]; ++k) {

                    //	Get the independent variable value.
                    value = TRUtils.nextNumber(tokenizer);
                    f64Value = Float64.valueOf(value);
                    if (i == 0 && j == 0)
                        indepArr3.add(f64Value);
                    else if (!indepArr3.get(k).equals(f64Value))
                        throw new IOException(format(RESOURCES.getString("postIncIndepValuesErr"), tblName, tokenizer.lineno()));

                    //	Get the dependent value.
                    value = TRUtils.nextNumber(tokenizer);
                    rank.add(Float64.valueOf(value));
                }
                depArray.get(i).add(Float64Vector.valueOf(rank));
                FastTable.recycle(rank);
            }
        }

        //	Create the table.
        Float64Table table = Float64Table.valueOf(tblName, indepNames, Float64Vector.valueOf(indepArr1),
                Float64Vector.valueOf(indepArr2), Float64Vector.valueOf(indepArr3), depArray);

        //	Clean up before leaving.
        FastTable.recycle(indepArr1);
        FastTable.recycle(indepArr2);
        FastTable.recycle(indepArr3);
        FastTable.recycle(depArray);

        return table;
    }

    /**
     * Write out the file level comments to the specified writer.
     */
    private static void writeFileComments(BufferedWriter output, FTableList tables) throws IOException {
        List<Text> notes = tables.getAllNotes();
        for (Text note : notes) {
            output.write("c ");
            note.println(output);
        }

        //	Write out some generic table comments.
        output.write("c Date and time written: ");
        output.write(DateFormat.getDateTimeInstance().format(new Date()));
        output.newLine();
        output.write("c");
        output.newLine();
    }

    /**
     * Write out a single table to a POST formatted output stream. This method assumes
     * that the "p$tab" line has already been output.
     *
     * @param output Reference to a BufferedWriter to write the specified table to.
     * @param table  The table to be written out.
     */
    private static void writeTable(BufferedWriter output, Float64Table table) throws IOException {

        //	Write out the table description.
        output.write(" table = '");
        Text.valueOf(table.getName()).print(output);
        output.write("', ");

        int numDim = table.getDimensions();
        TypeFormat.format(numDim, output);
        output.write(", ");
        for (int i = numDim - 1; i >= 0; --i) {
            Text varName = table.getIndepName(i);
            output.write("'");
            varName.print(output);
            output.write("', ");
        }

        for (int i = numDim - 1; i >= 0; --i) {
            int numBP = table.getNumBreakpoints(i);
            TypeFormat.format(numBP, output);
            output.write(", ");
        }

        output.write("8*1,");
        output.newLine();

        //	Write out the table based on the number of dimensions.
        switch (numDim) {
            case 1:
                write1DTable(output, table);
                break;
            case 2:
                write2DTable(output, table);
                break;
            case 3:
                write3DTable(output, table);
                break;
        }

    }

    /**
     * Write a 1D table to a POST table formatted output stream.
     *
     * @param output Reference to a BufferedWriter to write the specified table to.
     * @param table  The table to be written out.
     */
    private static void write1DTable(BufferedWriter output, Float64Table table) throws IOException {

        //	Get the breakpoints.
        Float64Vector indepArr = table.getBreakpoints(0);

        //	Loop over all the data and write them out.
        int numIndep = indepArr.getDimension();
        for (int i = 0; i < numIndep; ++i) {
            output.write("          ");
            formatNumber(8, indepArr.getValue(i)).print(output);
            output.write(",  ");
            formatNumber(16, table.getValue(i)).print(output);
            output.write(",");
            output.newLine();
        }

    }

    /**
     * Write a 2D table to a POST table formatted output stream.
     *
     * @param output Reference to a BufferedWriter to write the specified table to.
     * @param table  The table to be written out.
     */
    private static void write2DTable(BufferedWriter output, Float64Table table) throws IOException {

        //	Get the breakpoints.
        Float64Vector indepArr1 = table.getBreakpoints(0);
        Float64Vector indepArr2 = table.getBreakpoints(1);

        //	Loop over all the 1st independent variable.
        FastTable<Integer64> pos = FastTable.newInstance();
        pos.add(Integer64.ZERO);
        pos.add(Integer64.ZERO);
        int numIndep1 = indepArr1.getDimension();
        int numIndep2 = indepArr2.getDimension();
        for (int i = 0; i < numIndep1; ++i) {
            pos.set(0, Integer64.valueOf(i));		//	pos[0] = i;

            //	Output this independent
            output.write("      ");
            formatNumber(8, indepArr1.getValue(i)).print(output);
            output.write(",");
            output.newLine();

            //	Loop over the 2nd independent variable.
            for (int j = 0; j < numIndep2; ++j) {
                pos.set(1, Integer64.valueOf(j));	//	pos[1] = j;

                output.write("          ");
                formatNumber(8, indepArr2.getValue(j)).print(output);
                output.write(",  ");
                formatNumber(16, table.getValue(pos)).print(output);
                output.write(",");
                output.newLine();
            }
        }
        FastTable.recycle(pos);
    }

    /**
     * Write a 3D table to a POST table formatted output stream.
     *
     * @param output Reference to a BufferedWriter to write the specified table to.
     * @param table  The table to be written out.
     */
    private static void write3DTable(BufferedWriter output, Float64Table table) throws IOException {

        //	Get the breakpoints.
        Float64Vector indepArr1 = table.getBreakpoints(0);
        Float64Vector indepArr2 = table.getBreakpoints(1);
        Float64Vector indepArr3 = table.getBreakpoints(2);

        //	Loop over all the 1st independent variable.
        FastTable<Integer64> pos = FastTable.newInstance();
        for (int i = 0; i < 3; ++i)
            pos.add(Integer64.ZERO);
        int numIndep1 = indepArr1.getDimension();
        int numIndep2 = indepArr2.getDimension();
        int numIndep3 = indepArr3.getDimension();
        for (int i = 0; i < numIndep1; ++i) {
            pos.set(0, Integer64.valueOf(i));		//	pos[0] = i;

            //	Output this independent
            output.write("      ");
            formatNumber(8, indepArr1.getValue(i)).print(output);
            output.write(",");
            output.newLine();

            //	Loop over the 2nd independent variable.
            for (int j = 0; j < numIndep2; ++j) {
                pos.set(1, Integer64.valueOf(j));	//	pos[1] = j;

                //	Output this independent
                output.write("      ");
                formatNumber(8, indepArr2.getValue(j)).print(output);
                output.write(",");
                output.newLine();

                //	Loop over the 3rd independent variable.
                for (int k = 0; k < numIndep3; ++k) {
                    pos.set(2, Integer64.valueOf(k));	//	pos[2] = k;

                    output.write("          ");
                    formatNumber(8, indepArr3.getValue(k)).print(output);
                    output.write(",  ");
                    formatNumber(16, table.getValue(pos)).print(output);
                    output.write(",");
                    output.newLine();
                }
            }
        }
        FastTable.recycle(pos);
    }

    /**
     * Create a StreamTokenizer using the specified data and set it up properly.
     */
    private StreamTokenizer createTokenizer(String data) {

        //	Convert the String into a stream tokenizer.
        StreamTokenizer tokenizer = new StreamTokenizer(new StringReader(data));
        tokenizer.lowerCaseMode(true);
        tokenizer.commentChar('/');
        tokenizer.whitespaceChars(',', ',');
        tokenizer.wordChars('_', '_');
        tokenizer.wordChars('$', '$');
        tokenizer.wordChars('*', '*');
        tokenizer.wordChars('+', '+');
        tokenizer.wordChars('-', '-');

        return tokenizer;
    }

    /**
     * Takes the input string and strips out any lines that start with "C" or "c" in the
     * 1st column. Also replaces any characters from the 1st column with a space. Also
     * expands and repeated numbers (e.g.: "8*1" becomes "1, 1, 1, 1, 1, 1, 1, 1,".
     */
    private String preProcessesData(String data) {
        if (data.length() == 0)
            return data;
        StringBuilder buf = new StringBuilder(data.toLowerCase());

        //	First look for a "c" in the 1st column.
        int start = 0;
        while (true) {
            if (start != 0 || buf.charAt(0) != 'c') {
                start = buf.indexOf("\nc", start);
                if (start < 0)
                    break;
            }
            int end = buf.indexOf("\n", start + 1) + 1;
            if (end < 0)
                end = buf.length();
            buf = buf.delete(start, end);
        }

        //	Now replace any characters in the 1st column with spaces.
        start = -1;
        while (true) {
            if (start != -1) {
                start = buf.indexOf("\n", start);
                if (start < 0)
                    break;
            }
            ++start;
            if (start < buf.length())
                buf.setCharAt(start, ' ');
        }

        //	Now expand any repeated numbers
        int idx = -1;
        while (true) {
            idx = buf.indexOf("*", ++idx);
            if (idx < 0)
                break;

            //	Back up until white space or a comma is encountered.
            start = idx;
            char c = 0;
            while (c != ' ' && c != ',' && start > 0)
                c = buf.charAt(--start);
            ++start;

            //	Extract the repeat number.
            String str = buf.substring(start, idx);
            int repeat;
            try {
                //	Try to convert the string to a number.
                repeat = Integer.valueOf(str);
            } catch (Exception e) {
                //	On any exception, just continue on.
                continue;
            }

            //	Go forward until the next white space or comma.
            int end = idx;
            c = 0;
            while (c != ' ' && c != ',' && end < buf.length() - 1)
                c = buf.charAt(++end);

            //	Extract the number being repeated.
            str = buf.substring(idx + 1, end);
            int value;
            try {
                //	Try to convert the string to a number.
                value = Integer.valueOf(str);
            } catch (Exception e) {
                //	On any exception, just continue on.
                continue;
            }

            //	Delete the repeat coded number.
            buf = buf.delete(start, end);

            //	Now insert the repeated number.
            str = Integer.toString(value);
            buf = buf.insert(start, str);
            str = str + ",";
            for (int i = 1; i < repeat; ++i)
                buf = buf.insert(start, str);
        }

        //	Convert any Hollerith constants into quoted strings.
        idx = 0;
        while (true) {
            idx = buf.indexOf("h", ++idx);
            if (idx < 0)
                break;

            //	Back up until white space, equal sign, or a comma is encountered.
            start = idx;
            char c = 0;
            while (c != ' ' && c != ',' && c != '=' && start > 0)
                c = buf.charAt(--start);
            ++start;

            //	Extract the size number.
            String str = buf.substring(start, idx);
            int size;
            try {
                //	Try to convert the string to a number.
                size = Integer.valueOf(str);
            } catch (Exception e) {
                //	On any exception, just continue on looking for the next "h".
                continue;
            }

            //	Extract the string.
            int end = idx + 1 + size;
            str = buf.substring(idx + 1, end);

            //	Delete the Hollerith constant.
            buf = buf.delete(start, end);

            //	Insert a quoted string.
            buf = buf.insert(start, "'" + str + "'");
        }

        return buf.toString();
    }

    /**
     * Look for the "$tab table" sequence indicating the start of a table. If a "$tab
     * table" sequence is found, this method returns TT_WORD.
     */
    private int findTable(StreamTokenizer tokenizer) throws IOException {

        int ttype = tokenizer.ttype;
        while (ttype != TT_EOF) {
            if (ttype == TT_WORD && "$tab".equals(tokenizer.sval)) {
                //	Found "$tab", now, the next token needs to be "table".
                ttype = tokenizer.nextToken();
                if (ttype == TT_WORD && "table".equals(tokenizer.sval))
                    break;
            }
            ttype = tokenizer.nextToken();
        }

        return ttype;
    }

    /**
     * Format a number for output to a table file. This method formats the given number
     * using the a TextBuilder It then adds spaces to the start of the formatted number
     * until the string reaches the specified length.
     *
     * @param size   The overall length of the formatted number including the decimal
     *               point, minus sign, "E" notation, etc.
     * @param number The number to be formatted.
     */
    private static Text formatNumber(int size, double number) throws IOException {

        TextBuilder buffer = TextBuilder.newInstance();
        TypeFormat.format(number, size, false, false, buffer);
        int length = buffer.length();
        while (length < size) {
            buffer.insert(0, SPACE);
            ++length;
        }
        Text text = buffer.toText();
        TextBuilder.recycle(buffer);

        return text;
    }

}
